#define ELPP_STL_LOGGING

#include "hqs.hpp"

#include <iostream>
#include <vector>

#include <aigpp/EdgeRef.hh>
#include <easylogging/easylogging++.hpp>


namespace hqs {

std::ostream&
operator<<(std::ostream& stream, const Statistics& stat)
{
    stream << "Eliminated pure variables: " << stat.num_pure << '\n'
           << "Expanded universal variables (Henkin): " << stat.num_henkin_quantified << '\n'
           << "Eliminated existential variables: " << stat.num_exist_quantified << '\n'
           << "Eliminated dependencies: " << stat.num_dep_elim << '\n';
    return stream;
}

std::ostream&
operator<<(std::ostream& stream, const Settings& settings)
{
    stream << "Unit/pure checks: " << settings.unit_pure_checks << '\n'
           << "Filter:           " << settings.filter << '\n'
           << "FRAIGing:         " << settings.fraig << '\n'
           << "AIG rewriting:    " << settings.rewrite << '\n';
    return stream;
}

/**
 * \brief Constructs an DQBF solver object from a CNF representation.
 *
 * The matrix of the formula is converted into an AIG, where all gates that have been
 * recognized are substituted with their defining function. For proper DQBFs,
 * an elimination set is determined.
 *
 * \note This function modifies the parameter 'formula'. It may/should be freed afterwards.
 */
HenkinSolver::HenkinSolver(hqspre::Formula& formula) : _prefix(nullptr), _qbf_prefix(nullptr), _dqbf_prefix(nullptr)
{
    _manager.toggleAutomaticFRAIGing(false);

    // Check if we need an elimination set.
    if (!formula.isQBF()) {
        VLOG(1) << "Formula is a proper DQBF. Computing elimination set ... ";
        try {
            _dep_elim_set = formula.computeDepElimSet(false);
        } catch (hqspre::ElimSetException& e) {
            VLOG(2) << "Could not find a cost-optimal elimination set. Computing a size-minimal one.";
            _dep_elim_set = formula.computeDepElimSet(true);
        }
        formula.determineGates();
    } else {
        VLOG(1) << "Formula is a QBF. No elimination set needed.";
        formula.determineGates();
    }


    // Copy the formula's prefix
    VLOG(1) << "Creating quantifier prefix ... ";
    if (formula.getQBFPrefix() != nullptr) {
        _qbf_prefix = new hqspre::QBFPrefix(*formula.getQBFPrefix());
        _prefix     = _qbf_prefix;
    } else if (formula.getDQBFPrefix() != nullptr) {
        _dqbf_prefix = new hqspre::DQBFPrefix(*formula.getDQBFPrefix());
        _prefix      = _dqbf_prefix;
    }

    // Create AIG variables and fill _free_var_nums.
    for (Variable var = _prefix->minVarIndex(); var <= _prefix->maxVarIndex(); ++var) {
        if (_prefix->varDeleted(var)) {
            _free_var_nums.push_back(var);
        } else {
            _var2aig[var]                           = _manager.addVariable();
            _aig2var[_var2aig[var].variableIndex()] = var;
        }
    }

    // Convert the matrix into an AIG.
    VLOG(1) << "Converting matrix into an AIG ...";
    _matrix = createMatrix(formula);
    if (_settings.fraig) {
        _manager.makeFRAIG(10);
    }
    if (_settings.rewrite) {
        _manager.rewrite();
    }

    // Print some information about what needs to be done.
    VLOG(2) << "AIG construction finished.\n"
            << "  * number of variables: " << _var2aig.size() << '\n'
            << "  * size of the AIG: " << _matrix.nodeCount() << '\n'
            << "  * formula is a QBF: " << formula.isQBF() << '\n';
    if (!_dep_elim_set.empty() && VLOG_IS_ON(2)) {
        VLOG(2) << "Dependencies to eliminate:";
        for (Variable exist = _prefix->minVarIndex(); exist <= _prefix->maxVarIndex(); ++exist) {
            if (_prefix->isExistential(exist) && !_dep_elim_set[exist].empty()) {
                VLOG(2) << "     [" << exist << "] -> " << _dep_elim_set[exist];
            }
        }
    }

}

/**
 * \brief Destructor. Frees all memory used by the HenkinSolver object.
 */
HenkinSolver::~HenkinSolver()
{
    _var2aig.clear();
    _matrix.invalidate();
    delete _prefix;
}

/**
 * \brief Deletes a variable from the prefix.
 *
 * It's the user's responsibility that the AIGs no not depend
 * on the deleted variable anymore.
 */
void
HenkinSolver::removeVar(const Variable var)
{
    val_assert(!_prefix->varDeleted(var));
    val_assert(_var2aig[var].isValid());

    _prefix->removeVar(var);
    _aig2var.erase(_var2aig[var].variableIndex());
    _var2aig.erase(var);
    _free_var_nums.push_back(var);
}

/**
 * \brief Creates a copy of the given variable with the same dependencies.
 * \return the index of the created copy.
 */
Variable
HenkinSolver::copyVar(const Variable var)
{
    val_assert(_prefix->minVarIndex() <= var && var <= _prefix->maxVarIndex());
    val_assert(!_prefix->varDeleted(var));

    Variable new_var = 0;
    if (!_free_var_nums.empty()) {
        new_var = _free_var_nums.back();
        _free_var_nums.pop_back();
        val_assert(_prefix->varDeleted(new_var));
    } else {
        new_var = _prefix->maxVarIndex() + 1;
    }
    _prefix->makeCopy(new_var, var);
    _var2aig[new_var]                           = _manager.addVariable();
    _aig2var[_var2aig[new_var].variableIndex()] = new_var;
    return new_var;
}

bool
HenkinSolver::checkConsistency() const
{
    if (_prefix == nullptr) {
        LOG(ERROR) << "Prefix is null.";
        return false;
    }
    if ((_qbf_prefix == nullptr) == (_dqbf_prefix == nullptr)) {
        LOG(ERROR) << "Either _qbf_prefix or _dqbf_prefix must be null, but not both.";
        return false;
    }

    return _prefix->checkConsistency();
}

inline static aigpp::EdgeRef
convertLiteral(const Literal lit, const std::vector<aigpp::EdgeRef>& gate_table)
{
    const Variable var = hqspre::lit2var(lit);
    val_assert(var < gate_table.size());
    val_assert(gate_table[var].isValid());

    return gate_table[var].notIf(hqspre::isNegative(lit));
}

/**
 * \brief Converts a gate in CNF to its corresponding AIG representation
 */
aigpp::EdgeRef
HenkinSolver::gate2aig(const hqspre::Gate& g, const hqspre::Formula& preprocessor,
                       const std::vector<aigpp::EdgeRef>& gate_table)
{
    aigpp::EdgeRef result;

    switch (g._type) {
        case hqspre::GateType::AND_GATE: {
            if (g._input_literals.size() == 2) {
                result = aigpp::And(convertLiteral(g._input_literals[0], gate_table),
                                    convertLiteral(g._input_literals[1], gate_table));
            } else if (g._input_literals.size() == 1) {
                result = convertLiteral(g._input_literals[0], gate_table);
            } else {
                _aig_buffer.clear();
                _aig_buffer.reserve(g._input_literals.size());
                std::transform(g._input_literals.cbegin(), g._input_literals.cend(), std::back_inserter(_aig_buffer),
                               [&gate_table](Literal lit) { return convertLiteral(lit, gate_table); });
                result = aigpp::multiAnd(_manager, _aig_buffer);
            }
            break;
        }

        case hqspre::GateType::XOR_GATE: {
            result = aigpp::Xor(convertLiteral(g._input_literals[0], gate_table),
                                convertLiteral(g._input_literals[1], gate_table));
            break;
        }

        case hqspre::GateType::MUX_GATE: {
            result = aigpp::Mux(convertLiteral(g._input_literals[0], gate_table),   // select
                                convertLiteral(g._input_literals[1], gate_table),   // input 1
                                convertLiteral(g._input_literals[2], gate_table));  // input 0
            break;
        }

        case hqspre::GateType::UNKNOWN: {
            val_assert(hqspre::isPositive(g._output_literal));

            // determine the size of clauses containing the output positively/negatively
            unsigned int size_pos = 0;
            unsigned int size_neg = 0;
            for (const hqspre::ClauseID c_nr : g._encoding_clauses) {
                const hqspre::Clause& c = preprocessor.getClause(c_nr);
                if (c.containsLiteral(g._output_literal)) {
                    size_pos += c.size();
                } else {
                    size_neg += c.size();
                }
            }

            _aig_buffer.clear();
            std::vector<aigpp::EdgeRef> current_formula;

            if (size_pos < size_neg) {
                // Take the smaller set of clauses, hoping that it leads to a smaller AIG representation
                for (const hqspre::ClauseID c_nr : g._encoding_clauses) {
                    const hqspre::Clause& c = preprocessor.getClause(c_nr);
                    if (!c.containsLiteral(g._output_literal)) {
                        continue;
                    }

                    for (const Literal lit : c) {
                        if (lit == g._output_literal) {
                            continue;
                        }
                        _aig_buffer.push_back(convertLiteral(lit, gate_table));
                    }
                    current_formula.push_back(aigpp::multiOr(_manager, _aig_buffer));
                    _aig_buffer.clear();
                }
                result = aigpp::multiAnd(_manager, current_formula);
            } else {
                for (const hqspre::ClauseID c_nr : g._encoding_clauses) {
                    const hqspre::Clause& c = preprocessor.getClause(c_nr);
                    if (c.containsLiteral(g._output_literal)) {
                        continue;
                    }

                    for (const Literal lit : c) {
                        if (lit == hqspre::negate(g._output_literal)) {
                            continue;
                        }
                        _aig_buffer.push_back(convertLiteral(lit, gate_table));
                    }
                    current_formula.push_back(aigpp::multiOr(_manager, _aig_buffer));
                    _aig_buffer.clear();
                }
                result = aigpp::Not(aigpp::multiAnd(_manager, current_formula));
            }

            break;
        }

        default: {
            val_assert_msg(false, "Invalid gate type encountered!");
            break;
        }
    }

    val_assert(result.isValid());

    return result.notIf(hqspre::isNegative(g._output_literal));
}

aigpp::EdgeRef
HenkinSolver::createMatrix(hqspre::Formula& formula)
{
    const auto gates = formula.getGates();

    std::vector<aigpp::EdgeRef> gate_table(formula.maxVarIndex() + 1, aigpp::EdgeRef());
    val_assert(!gate_table[0].isValid());

    // Load the gate_table with the primary inputs
    for (Variable var = formula.minVarIndex(); var <= formula.maxVarIndex(); ++var) {
        if (!formula.varDeleted(var) && !formula.isGateOutput(var)) {
            gate_table[var] = var2aig(var);
        }
    }

    // Create the gates
    for (const hqspre::Gate& g : gates) {
        const Variable out_var = hqspre::lit2var(g._output_literal);
        gate_table[out_var]    = gate2aig(g, formula, gate_table);
        for (const auto c_nr : g._encoding_clauses) {
            formula.removeClause(c_nr);
        }
        removeVar(out_var);
    }

    // Convert the clauses
    std::vector<aigpp::EdgeRef> clause_aigs;
    clause_aigs.reserve(formula.numClauses());
    _aig_buffer.clear();

    for (hqspre::ClauseID c_nr = 0; c_nr <= formula.maxClauseIndex(); c_nr++) {
        if (!formula.clauseDeleted(c_nr)) {
            const auto& clause = formula.getClause(c_nr);
            _aig_buffer.reserve(clause.size());
            for (const Literal lit : clause) {
                _aig_buffer.push_back(convertLiteral(lit, gate_table));
                val_assert(_aig_buffer.back().isValid());
            }
            clause_aigs.push_back(aigpp::multiOr(_manager, _aig_buffer));
            _aig_buffer.clear();
        }
    }

    return aigpp::multiAnd(_manager, clause_aigs);
}

/**
 * \brief Removes a universal variable from the formula 'aig' by
 *        universal quantification.
 *
 * \param[out] aig the formula from which the variable is to be
 *               eliminated; this parameter is modified by quantification.
 * \param[in] var the variable to be quantified.
 * \return TRUE/FALSE if the truth value of the formula could be determined
 *          during quantification, UNKNOWN otherwise.
 * \sa Henkinsolver::quantifyExist(aigpp::EdgeRef&)
 */
TruthValue
HenkinSolver::quantifyHenkin(aigpp::EdgeRef& aig, const Variable var)
{
    // Check for consistency
    val_assert(_dqbf_prefix != nullptr);
    val_assert(1 <= var && var <= _prefix->maxVarIndex());
    // std::cout << "Expanding " << var << '\n';
    if (_prefix->varDeleted(var)) {
        // The variable has already been eliminated. So there is no need to
        // eliminate it now ...
        return TruthValue::UNKNOWN;
    }

    val_assert(_prefix->isUniversal(var));
    val_assert(checkConsistency());

    // Compute cofactors
    aigpp::EdgeRef cof1 = aig.cofactor(var2aig(var));
    aigpp::EdgeRef cof0 = aig.cofactor(!var2aig(var));

    if (_settings.unit_pure_checks) {
        const auto pure0 = findPureVariables(cof0, _prefix->maxVarIndex());
        const auto pure1 = findPureVariables(cof1, _prefix->maxVarIndex());

        for (Variable y = _prefix->minVarIndex(); y <= _prefix->maxVarIndex(); ++y) {
            if (_prefix->varDeleted(y)) {
                continue;
            }
            const bool dep = _prefix->isExistential(y) && _prefix->depends(var, y);
            if (dep) {
                if (pure0[y] == PureStatus::POSITIVE) {
                    cof0 = cof0.cofactor(var2aig(y));
                    _dqbf_prefix->removeDependency(var, y);
                    ++_statistics.num_pure;
                } else if (pure0[y] == PureStatus::NEGATIVE) {
                    cof0 = cof0.cofactor(!var2aig(y));
                    _dqbf_prefix->removeDependency(var, y);
                    ++_statistics.num_pure;
                } else if (pure1[y] == PureStatus::POSITIVE) {
                    cof1 = cof1.cofactor(var2aig(y));
                    _dqbf_prefix->removeDependency(var, y);
                    ++_statistics.num_pure;
                } else if (pure1[y] == PureStatus::NEGATIVE) {
                    cof1 = cof1.cofactor(!var2aig(y));
                    _dqbf_prefix->removeDependency(var, y);
                    ++_statistics.num_pure;
                } else if ((pure0[y] == PureStatus::IRRELEVANT || pure1[y] == PureStatus::IRRELEVANT)) {
                    ++_statistics.num_pure;
                    _dqbf_prefix->removeDependency(var, y);
                }
            } else if (_prefix->isExistential(y)) {
                // var and y do not depend on each other
                if (pure0[y] == PureStatus::POSITIVE && pure0[y] == pure1[y]) {
                    cof0 = cof0.cofactor(var2aig(y));
                    cof1 = cof1.cofactor(var2aig(y));
                    ++_statistics.num_pure;
                } else if (pure0[y] == PureStatus::NEGATIVE && pure0[y] == pure1[y]) {
                    cof0 = cof0.cofactor(!var2aig(y));
                    cof1 = cof1.cofactor(!var2aig(y));
                    ++_statistics.num_pure;
                }
            }
        }
    }  // end if (unit_pure_checks)

    const auto deps = _dqbf_prefix->getDependencies(var);

    // Create copies of all variables which depend on `var`.
    _aig_buffer.clear();
    const std::size_t            num_deps = deps.size();
    std::vector<aigpp::EdgeRef>& old_vars = _aig_buffer;
    std::vector<aigpp::EdgeRef>  new_vars;
    old_vars.reserve(num_deps);
    new_vars.reserve(num_deps);

    for (const Variable orig_var : deps) {
        const Variable new_var = copyVar(orig_var);
        old_vars.push_back(var2aig(orig_var));
        new_vars.push_back(var2aig(new_var));
    }

    // Remove the quantified variable from all lists.
    _prefix->removeVar(var);
    _var2aig.erase(var);

    // Replace in one cofactor (here: cofactor 1) the dependent variables
    // by their copies
    cof1 = cof1.compose(old_vars, new_vars);
    aig  = aigpp::And(cof0, cof1);

    // Do AIG simplification (FRAIGing not necessary as this is already
    // done by the quantification routine.
    cof0.invalidate();
    cof1.invalidate();
    if (_settings.rewrite) {
        _manager.rewrite();
    }

    TruthValue result = TruthValue::UNKNOWN;
    if (aigpp::functionallyTrue(aig)) {
        result = TruthValue::TRUE;
    } else if (aigpp::functionallyFalse(aig)) {
        result = TruthValue::FALSE;
    }

    ++_statistics.num_henkin_quantified;

    val_assert(checkConsistency());

    return result;
}

/**
 * \brief Eliminates a dependency by replacing the existential variable \f$y\f$ by \f$x\land y^1\lor \neg x\land y^0\f$.
 * \param root the AIG in which the variable dependency should be eliminated
 * \param exist_var the existential variable
 * \param univ_var the universal variable
 * \return the two existential variables representing the dependency, i. e., \f$(y^0, y^1)\f$.
 * \note the returned variable indicies are not necessarily disjoint from <tt>exist_var</tt>. If
 *       <tt>exist_var</tt> does not depend on <tt>univ_var</t>>, the function returns \f$(0,0)\f$.
 */
std::pair<Variable, Variable>
HenkinSolver::eliminateDependency(aigpp::EdgeRef& root, const Variable univ_var, const Variable exist_var)
{
    val_assert(_dqbf_prefix != nullptr);

    if (_prefix->varDeleted(exist_var) || _prefix->varDeleted(univ_var)) {
        // One of the variables has already been eliminated. So there is no need to
        // eliminate the dependency
        return std::make_pair(0, 0);
    }

    val_assert(_prefix->isExistential(exist_var));
    val_assert(_prefix->isUniversal(univ_var));
    val_assert(exist_var <= _prefix->maxVarIndex());
    val_assert(univ_var <= _prefix->maxVarIndex());

    if (_prefix->depends(univ_var, exist_var)) {
        _dqbf_prefix->removeDependency(exist_var, univ_var);
    } else {
        return std::make_pair(0, 0);
    }

    const Variable new_var0 = exist_var;
    const Variable new_var1 = copyVar(exist_var);

    const auto replacement = aigpp::Mux(var2aig(univ_var), var2aig(new_var1), var2aig(new_var0));
    root                   = root.compose({var2aig(exist_var)}, {replacement});

    ++_statistics.num_dep_elim;

    return std::make_pair(new_var0, new_var1);
}

/**
 * \brief Solves the given DQBF.
 *
 * \return TruthValue::TRUE if the formula is satisfiable,
 *         TruthValue::FALSE if it is unsatisfiable, and
 *         TruthValue::UNKNOWN if the formula could not be solved
 *         due to time or memory limitations.
 */
TruthValue
HenkinSolver::solve()
{
    if (_dqbf_prefix != nullptr && !_dqbf_prefix->isQBF()) {
        quantifyExist(_matrix);

        // Call the QBF filtering procedure
        if (_settings.filter) {
            VLOG(1) << "Solving QBF filter ... ";
            const auto result = fastRefutation();
            if (result == TruthValue::FALSE) {
                VLOG(1) << "Solved by QBF filter.";
                return TruthValue::FALSE;
            }
        }

        if (_settings.eqDQBF) {
            VLOG(1) << "Localize quantifiers ...";
            const TruthValue result = earlyQuantifyDQBF();

            if (result != TruthValue::UNKNOWN) {
                VLOG(1) << "Solved by solver core (quantifier localization on DQBF)." << std::endl;
                return result;
            }
        }


        VLOG(1) << "Turning DQBF into QBF ...";
        const auto result = turnIntoQBF();
        if (result != TruthValue::UNKNOWN) {
            VLOG(1) << "Solved by solver core (DQBF to QBF phase).\n";
            return result;
        }
    }


    val_assert(_qbf_prefix != nullptr || (_dqbf_prefix != nullptr && _dqbf_prefix->isQBF()));

    VLOG(1) << "Solving QBF after conversion ...";

    const auto                  qbf_prefix = createQBFPrefix();
    std::vector<VariableStatus> quantifiers(_prefix->maxVarIndex() + 1, VariableStatus::DELETED);
    for (Variable var = _prefix->minVarIndex(); var <= _prefix->maxVarIndex(); ++var) {
        quantifiers[var] = _prefix->getStatus(var);
    }
    const TruthValue result = solveQBF(qbf_prefix, quantifiers, _matrix);

    VLOG_IF(result != TruthValue::UNKNOWN, 1) << "Solved by solver core (QBF solution phase).\n";
    return result;
}

std::vector<PureStatus>
HenkinSolver::findPureVariables(const aigpp::EdgeRef& formula, const std::size_t max_index)
{
    // Initialize the status of all existing variables with 'irrelevant'
    std::vector<PureStatus> pure_info(max_index + 1, PureStatus::IRRELEVANT);

    // Get the polarities of all variables
    aigpp::InternalEdgeRef                           aig_internal = formula.getInternal();
    const std::vector<std::pair<aigpp::Node*, bool>> polarities   = aig_internal.getVariablePolarities();

    for (const auto& aig_var : polarities) {
        val_assert(aig_var.first != nullptr);
        val_assert(aig_var.first->isVar());

        const Variable var_index = aig2var(aig_var.first->varIndex());

        if (aig_var.second) {
            // path with even polarity exists
            if (pure_info[var_index] == PureStatus::NEGATIVE) {
                pure_info[var_index] = PureStatus::NOTPURE;
            } else {
                pure_info[var_index] = PureStatus::POSITIVE;
            }
        } else {
            // path with odd polarity exists
            if (pure_info[var_index] == PureStatus::POSITIVE) {
                pure_info[var_index] = PureStatus::NOTPURE;
            } else {
                pure_info[var_index] = PureStatus::NEGATIVE;
            }
        }
    }

    return pure_info;
}

bool
HenkinSolver::eliminatePureVariables(std::vector<VariableStatus>& quantifiers, aigpp::EdgeRef& formula, bool remove)
{
    const std::vector<PureStatus> pureInfo = findPureVariables(formula, quantifiers.size() - 1);

    // Substitute the detected pure variables
    _aig_buffer.clear();
    std::vector<aigpp::EdgeRef>& pure_vars = _aig_buffer;
    pure_vars.reserve(pureInfo.size());
    std::size_t count = 0;

    for (Variable var_index = 1; var_index < quantifiers.size(); ++var_index) {
        if (quantifiers[var_index] == VariableStatus::DELETED) {
            continue;
        }

        switch (pureInfo[var_index]) {
            case PureStatus::IRRELEVANT:
                // Variable does not appear in the AIG. Remove it from all lists.
                if (remove) {
                    removeVar(var_index);
                }
                break;

            case PureStatus::NOTPURE:
                // nothing to do
                break;

            case PureStatus::POSITIVE:
                pure_vars.push_back(
                    lit2aig(hqspre::var2lit(var_index, quantifiers[var_index] == VariableStatus::UNIVERSAL)));
                if (remove) {
                    removeVar(var_index);
                }
                ++count;
                break;

            case PureStatus::NEGATIVE:
                pure_vars.push_back(
                    lit2aig(hqspre::var2lit(var_index, quantifiers[var_index] == VariableStatus::EXISTENTIAL)));
                if (remove) {
                    removeVar(var_index);
                }
                ++count;
                break;
        }
    }  // end for (pure)

    // Now we can perform the actual substitution
    if (!pure_vars.empty()) {
        formula = formula.cofactor(pure_vars);
    }

    _statistics.num_pure += count;

    VLOG_IF(count > 0, 2) << "Eliminated " << count << " pure variables.";
    return count > 0;
}

TruthValue
HenkinSolver::quantifyExist(aigpp::EdgeRef& formula)
{
    const auto rmb = _prefix->getRMB();
    if (rmb.empty()) {
        return TruthValue::UNKNOWN;
    }

    _aig_buffer.clear();
    _aig_buffer.reserve(rmb.size());

    for (const auto var : rmb) {
        _aig_buffer.push_back(var2aig(var));
        removeVar(var);
    }

    _manager.bddQuantifyInplace(formula, _aig_buffer, true, false, true);

    VLOG_IF(!_aig_buffer.empty(), 2) << "Eliminated " << _aig_buffer.size()
                                     << " existential variables. AIG size = " << formula.nodeCount();

    _statistics.num_exist_quantified += _aig_buffer.size();

    return TruthValue::UNKNOWN;
}

TruthValue
HenkinSolver::turnIntoQBF()
{
    if (_dqbf_prefix == nullptr || _dqbf_prefix->isQBF()) {
        return TruthValue::UNKNOWN;
    }
    val_assert(!_dep_elim_set.empty());

    // Find out which universal variables to expand and which dependencies to eliminate.
    std::vector<int> num_to_elim(_prefix->maxVarIndex() + 1, 0);
    for (Variable exist = _prefix->minVarIndex(); exist <= _prefix->maxVarIndex(); ++exist) {
        if (!_prefix->isExistential(exist)) {
            continue;
        }
        const auto& deps = _dqbf_prefix->getDependencies(exist);
        for (const Variable univ : deps) {
            ++num_to_elim[univ];
        }

        for (const Variable univ : _dep_elim_set[exist]) {
            --num_to_elim[univ];
        }
    }
    std::vector<Variable> current_vars;
    std::vector<Variable> next_vars;

    std::vector<Variable> to_expand;
    std::vector<bool>     seen(_prefix->maxVarIndex() + 1, false);

    for (Variable exist = _prefix->minVarIndex(); exist <= _prefix->maxVarIndex(); ++exist) {
        if (exist >= _dep_elim_set.size()) {
            break;
        }
        if (!_prefix->isExistential(exist) || _dep_elim_set[exist].empty()) {
            continue;
        }

        current_vars.push_back(exist);
        for (Variable univ : _dep_elim_set[exist]) {
            if (num_to_elim[univ] == 0 && !seen[univ]) {
                to_expand.push_back(univ);
                seen[univ] = true;
                continue;  // skip universal variables which are expanded later.
            } else if (seen[univ]) {
                continue;
            }

            VLOG(2) << "Eliminating dependency of " << exist << " on " << univ << " ...";
            for (Variable current_exist : current_vars) {
                const auto vars = eliminateDependency(_matrix, univ, current_exist);
                if (vars.first > 0) {
                    next_vars.push_back(vars.first);
                }
                if (vars.second > 0) {
                    next_vars.push_back(vars.second);
                }
            }
            current_vars.clear();
            std::swap(current_vars, next_vars);
        }
    }

    // Finally expand all universal variables of which all dependencies must be removed.
    std::sort(to_expand.begin(), to_expand.end(), [this](const Variable a, const Variable b) {
        return _dqbf_prefix->getDependencies(a).size() < _dqbf_prefix->getDependencies(b).size();
    });

    for (const auto univ : to_expand) {
        VLOG(2) << "Expanding universal variable " << univ << " ...";
        const auto result = quantifyHenkin(_matrix, univ);
        //        if (_settings.unit_pure_checks) {
        //            eliminatePureVariables();
        //        }
        //        if (_settings.rewrite) _manager.rewrite();
        if (result != TruthValue::UNKNOWN) {
            return result;
        }
    }

    return TruthValue::UNKNOWN;
}

std::vector<Variable>
HenkinSolver::createQBFPrefix()
{
    std::vector<Variable> order;
    order.reserve(_prefix->maxVarIndex());

    if (_qbf_prefix != nullptr) {
        for (unsigned int level = 0; level <= _qbf_prefix->getMaxLevel(); ++level) {
            const auto& block = _qbf_prefix->getVarBlock(level);
            std::copy(block.cbegin(), block.cend(), std::back_inserter(order));
        }
        return order;
    }

    val_assert(_dqbf_prefix != nullptr);

    // Get all existential variables
    std::vector<Variable> exist_vars;
    for (Variable var = _prefix->minVarIndex(); var <= _prefix->maxVarIndex(); ++var) {
        if (_prefix->isExistential(var)) {
            exist_vars.push_back(var);
        }
    }

    // Sort existential variables according to their number of dependencies.
    // This brings them into the right order for the QBF prefix.
    std::sort(exist_vars.begin(), exist_vars.end(), [this](Variable a, Variable b) {
        const auto deps_a = _prefix->numDependencies(a);
        const auto deps_b = _prefix->numDependencies(b);
        return (deps_a < deps_b) || (deps_a == deps_b && a < b);
    });

    // Insert the universal variables into the order.
    std::vector<bool> finished_universal(_prefix->maxVarIndex() + 1, false);
    for (const auto exist : exist_vars) {
        const auto& deps = _dqbf_prefix->getDependencies(exist);
        for (const auto dep : deps) {
            if (!finished_universal[dep]) {
                finished_universal[dep] = true;
                order.push_back(dep);
            }
        }
        order.push_back(exist);
    }

    // Append the remaining universal variables which do not appear in
    // any dependency sets.
    for (Variable univ = _prefix->minVarIndex(); univ <= _prefix->maxVarIndex(); ++univ) {
        if (_prefix->isUniversal(univ) && !finished_universal[univ]) {
            finished_universal[univ] = true;
            order.push_back(univ);
        }
    }

    return order;
}

TruthValue
HenkinSolver::solveQBF(const std::vector<Variable>& prefix, std::vector<VariableStatus>& quantifiers,
                       aigpp::EdgeRef formula)
{
    VariableStatus quantifier = quantifiers[prefix.back()];

    _aig_buffer.clear();
    _aig_buffer.reserve(prefix.size());

    if (_settings.unit_pure_checks) {
        eliminatePureVariables(quantifiers, formula, false);
    }

    for (auto it = prefix.crbegin(); it != prefix.crend(); ++it) {
        const Variable       var    = *it;
        const VariableStatus status = quantifiers[var];
        if (status == VariableStatus::DELETED) {
            continue;
        }
        if (quantifier == VariableStatus::DELETED) {
            quantifier = status;
        }

        if (status == quantifier) {
            _aig_buffer.push_back(var2aig(var));
        } else {
            if (!_aig_buffer.empty()) {
                VLOG(2) << "Quantifying a block of " << _aig_buffer.size() << " "
                        << (quantifier == VariableStatus::EXISTENTIAL ? "existential" : "universal")
                        << " variables ... ";
                _manager.bddQuantifyInplace(formula, _aig_buffer, quantifier == VariableStatus::EXISTENTIAL, false,
                                            true);
                if (_settings.fraig) {
                    _manager.makeFRAIG(10);
                }
                if (_settings.rewrite) {
                    _manager.rewrite();
                }
                _aig_buffer.clear();

                if (aigpp::functionallyFalse(formula)) {
                    return TruthValue::FALSE;
                } else if (aigpp::functionallyTrue(formula)) {
                    return TruthValue::TRUE;
                }
            }
            _aig_buffer.push_back(var2aig(var));
            quantifier = status;
            //            if (_settings.unit_pure_checks) {
            //                eliminatePureVariables(quantifiers, formula, false);
            //            }
        }
    }

    // Now only the outermost variable block is left.
    if (_aig_buffer.empty()) {
        if (aigpp::functionallyTrue(formula)) {
            return TruthValue::TRUE;
        } else {
            return TruthValue::FALSE;
        }
    } else if (quantifier == hqspre::VariableStatus::EXISTENTIAL) {
        _aig_buffer.clear();
        if (aigpp::functionallyFalse(formula)) {
            return TruthValue::FALSE;
        } else {
            return TruthValue::TRUE;
        }
    } else {
        _aig_buffer.clear();
        if (aigpp::functionallyTrue(formula)) {
            return TruthValue::TRUE;
        } else {
            return TruthValue::FALSE;
        }
    }

    _aig_buffer.clear();
    return TruthValue::UNKNOWN;
}

TruthValue
HenkinSolver::fastRefutation()
{
    val_assert(_dqbf_prefix != nullptr);

    aigpp::EdgeRef              formula = _matrix;
    const auto                  prefix  = createQBFPrefix();
    std::vector<VariableStatus> quantifiers(_prefix->maxVarIndex() + 1, VariableStatus::DELETED);
    for (Variable var = _prefix->minVarIndex(); var <= _prefix->maxVarIndex(); ++var) {
        quantifiers[var] = _prefix->getStatus(var);
    }

    if (solveQBF(prefix, quantifiers, formula) == TruthValue::FALSE) {
        return TruthValue::FALSE;
    } else {
        return TruthValue::UNKNOWN;
    }
}

}  // end namespace hqs
