#ifndef HQS_HPP
#    define HQS_HPP

#    include <iosfwd>
#    include <unordered_map>

#    include <aigpp/EdgeRef.hh>
#    include <aigpp/Manager.hh>
#    include <formula.hpp>
#    include <prefix.hpp>

namespace hqs {

using Variable       = hqspre::Variable;
using VariableStatus = hqspre::VariableStatus;
using Literal        = hqspre::Literal;
using TruthValue     = hqspre::TruthValue;

struct Settings
{
    bool unit_pure_checks = true;
    bool filter           = true;
    bool fraig            = false;
    bool rewrite          = true;
    bool eqDQBF           = true;
};

std::ostream& operator<<(std::ostream& stream, const Settings& settings);

struct Statistics
{
    unsigned int num_henkin_quantified = 0;
    unsigned int num_exist_quantified  = 0;
    unsigned int num_univ_quantified   = 0;
    unsigned int num_dep_elim          = 0;
    unsigned int num_pure              = 0;
};

std::ostream& operator<<(std::ostream& stream, const Statistics& stat);

// Status of a variable regarding unit and pure
enum class PureStatus
{
    POSITIVE,   // Variable is positive pure / positive unit
    NEGATIVE,   // Variable is negative pure / negative unit
    NOTPURE,    // Variable is not pure or unit
    IRRELEVANT  // AIG does not depend on this variable
};

class HenkinSolver
{
   public:
    explicit HenkinSolver(hqspre::Formula& formula);
    ~HenkinSolver();

    // Forbid copying and moving HenkinSolver objects
    explicit HenkinSolver(const HenkinSolver& other) = delete;
    explicit HenkinSolver(HenkinSolver&& other)      = delete;
    HenkinSolver& operator=(const HenkinSolver& other) = delete;
    HenkinSolver& operator=(HenkinSolver&& other) = delete;

    // Access to statistics and settings
    const Statistics& statistics() const noexcept;
    const Settings&   settings() const noexcept;
    Settings&         settings() noexcept;

    // Formula solving
    TruthValue solve();

   private:
    // Variable management
    Variable copyVar(Variable original);
    void     removeVar(Variable var);

    // Quantification routines
    TruthValue                    quantifyHenkin(aigpp::EdgeRef& aig, Variable var);
    TruthValue                    quantifyExist(aigpp::EdgeRef& aig);
    std::pair<Variable, Variable> eliminateDependency(aigpp::EdgeRef& aig, Variable univ, Variable exist);

    // Unit/pure checks
    std::vector<PureStatus> findPureVariables(const aigpp::EdgeRef& formula, const std::size_t max_index);
    bool eliminatePureVariables(std::vector<VariableStatus>& quantifiers, aigpp::EdgeRef& formula, bool remove);

    TruthValue solveQBF(const std::vector<Variable>& prefix, std::vector<VariableStatus>& quantifiers,
                        aigpp::EdgeRef formula);
    TruthValue turnIntoQBF();

    bool checkConsistency() const;

    // Variable/clause to AIG conversion
    aigpp::EdgeRef var2aig(Variable var) const;
    Variable       aig2var(const aigpp::EdgeRef& aig) const;
    Variable       aig2var(unsigned int var) const;
    aigpp::EdgeRef lit2aig(Literal lit) const;

    template <typename Container>
    aigpp::EdgeRef clause2aig(const Container& clause);

    aigpp::EdgeRef gate2aig(const hqspre::Gate& g, const hqspre::Formula& formula,
                            const std::vector<aigpp::EdgeRef>& gate_table);
    aigpp::EdgeRef createMatrix(hqspre::Formula& formula);

    // Fast refutation
    std::vector<Variable> createQBFPrefix();
    TruthValue            fastRefutation();

    // Data
    aigpp::Manager                               _manager;
    hqspre::Prefix*                              _prefix;
    hqspre::QBFPrefix*                           _qbf_prefix;
    hqspre::DQBFPrefix*                          _dqbf_prefix;
    aigpp::EdgeRef                               _matrix;
    Settings                                     _settings;
    Statistics                                   _statistics;
    std::unordered_map<Variable, aigpp::EdgeRef> _var2aig;
    std::unordered_map<unsigned int, Variable>   _aig2var;
    std::vector<Variable>                        _free_var_nums;
    std::vector<std::vector<Variable>>           _dep_elim_set;
    std::vector<aigpp::EdgeRef>                  _aig_buffer;




    // quantifier localization /////////////////
    // common for QBF and DQBF
    struct SimpleEdge;
    struct SimpleGate;

    // structure has root (output) on top and leaves (inputs) at the bottom.
    // each gate has incoming edges from above (children) and
    // exaclty two outgoing edges (parents)
    // children and parents are the edges
    // each edge (except from rootEdge) has an origin (source) and a myGate (target)
    struct SimpleGate
    {
        SimpleGate();
        ~SimpleGate();

        unsigned int id;
        unsigned int var;
        bool         isOR;  // 0 for AND and 1 for OR

        SimpleEdge*           parent1;
        SimpleEdge*           parent2;
        std::set<SimpleEdge*> children; // for rootgate this set is empty
        std::vector<bool> supports;      // for variable v, index v is true

        SimpleGate*           macroOrigin; // = stored in each gate from the macrogate
        std::set<SimpleEdge*> macroParents; // only stored for the root of the macrogate
    };

    struct SimpleEdge
    {
        SimpleEdge();
        ~SimpleEdge();

        bool                  isNeg;
        SimpleGate*           myGate;        // gate towards the inputs
        SimpleGate*           origin;        // gate towards the output
        std::vector<Variable> simplePrefix;  // first var is outermost one. for DQBF forall vars are first ones

        std::set<SimpleEdge*> prefixChildren;  // prefixes at the output edges
        std::set<SimpleEdge*> prefixParents;   // prefixes at input edges

        std::set<Variable> seenUnivVars;  // all universal variables in whose scope this edge is, including univ vars at this edge.
    };

    // variables and data structures for DQBF and QBF
    unsigned int maxVarIndex;
    unsigned int maxGateID;
    SimpleEdge*  rootEdge;
    SimpleGate*  rootGate;

    std::map<Variable, SimpleGate*> primInputs;
    std::map<Variable, std::set<SimpleEdge*>> parsWithVar;

    std::vector<aigpp::EdgeRef>               aigvec;
    std::map<unsigned int, aigpp::EdgeRef>    aigs;

    unsigned int cntQuantifications;

    // reused in graph traversal
    std::vector<unsigned int> seenChildren;  // cnt if we have seen all children already
    std::vector<bool>         seenParents;   // cnt if we have seen both parents already
    std::vector<bool>         seenGates;     // check if we have seen this gate already
    std::set<SimpleEdge*>     seenEdges;
    std::vector<bool>         seenVariables;

    std::vector<Variable> quantableVars;

    // functions usable for QBF and DQBF, all in hqs_ql.cpp
    void         initializeSimpleStructure();
    SimpleEdge*  getSimpleStructure();
    void         buildNNF();
    void         getSupports();
    bool         buildMacroGate(SimpleGate* gate);
    void         deleteVarFromChildren(SimpleGate* gate, Variable var);
    void         getParsWithVar(SimpleGate* gate, std::vector<Variable>& vars);

    void         clearSimpleStructure();  // TODO: check if this function really clears everything!!

    // quantifier localization on DQBF, all in hqs_ql_dqbf.cpp
    // TODO: check which functions can be reused in QBF case.
    TruthValue   earlyQuantifyDQBF();
    bool         cleanUpUnnecessaryVars();
    bool         localizeQuantifiersDQBF();
    void         computeCommonVarsDQBF(SimpleEdge* edge, std::vector<Variable>& commonVars);
    bool         pushVarsDQBF(SimpleEdge* edge, std::vector<Variable>& commonVars);
    bool         pushVarsMacroDQBF(SimpleEdge* edge, std::vector<Variable>& commonVars);

    bool         pushVarsMacroAnd(SimpleGate* gate, std::vector<Variable>& commonVars);
    bool         pushVarsMacroOr(SimpleGate* gate, std::vector<Variable>& commonVars);
    unsigned int findSmallestVarAND(std::vector<Variable>& vars);
    unsigned int findSmallestVarOR(SimpleGate* gate, std::vector<Variable>& vars);
    bool isPushableUniv(Variable var);
    unsigned int collectParentsOR(SimpleGate* gate, Variable var);
    bool         isPushableEx(const std::set<SimpleEdge*>& parents, Variable exVar, SimpleGate* gate);
    void         collectCone(SimpleGate* gate, std::set<unsigned int>& cone);

    void         findPrefixTopologyAndOccurrences(
                    SimpleEdge* edge, std::vector<std::set<SimpleEdge*>>& varAtEdges);
    void         deleteDeps(std::vector<std::set<SimpleEdge*>>& varAtEdges);
    void         buildEarlyQuantifiedDQBF(aigpp::EdgeRef& newRootEdgeRef,
                    std::vector<std::set<SimpleEdge*>>& varAtEdges);

    void         distributeVarsMacro(SimpleGate* gate, Variable var);
    void         restructureMacroParents(SimpleGate* gate, const unsigned int var);
    void         restructureMacroParentsUnivOR(SimpleGate* gate, Variable var);

    void         findTopoOrder(SimpleGate* gate, std::vector<SimpleGate*>& orderedGates);
    void         swapParents(SimpleEdge* parentStayAtOutGate, SimpleEdge* parentSwapToOutGate);
    void         updateSupports(SimpleGate* gate);
    void         addDepPars(SimpleGate* gate, Variable var, std::set<SimpleEdge*>& parsInclDeps);

    void         adjustSupports(const std::vector<SimpleGate*>& gates);

    void         buildAIG(SimpleGate* gate, std::vector<std::set<SimpleEdge*>>& varAtEdges);
    void         quantify(SimpleEdge* edge, aigpp::EdgeRef& edgeRef, std::vector<std::set<SimpleEdge*>>& varAtEdges);

    bool         isQuantableE(SimpleGate* gate, Variable var);
    bool         isQuantableA(SimpleEdge* edge, Variable var);
    void         quantGate(std::vector<Variable>& quantVars, aigpp::EdgeRef& edgeRef);
    void         pushVarsToChildren(SimpleEdge* edge, std::vector<std::set<SimpleEdge*>>& varAtEdges);

    // new approach to split nodes to enable pushing
    bool pushVarsMacroDQBF2(SimpleEdge* edge, std::vector<SimpleGate*>& newGates);
    bool pushVarsDQBF2(SimpleEdge* edge, std::vector<SimpleGate*>& newGates);
    void getVarsAtChildren(SimpleGate* gate, std::map<Variable,std::set<SimpleEdge*>>& varsAtChildren/*, std::vector<Variable>& occurringVars*/);
    SimpleGate* splitGate(SimpleGate* origGate, std::set<SimpleEdge*>& childrenForNewGate);
    SimpleGate* splitMacroGate(SimpleGate* origGate, std::set<SimpleEdge*>& childrenForNewGate);
    bool pushVarsMacroDQBFORSplit(SimpleGate* gate, std::vector<SimpleGate*>& newGates);
    bool pushVarsMacroDQBFANDSplit(SimpleGate* gate, std::vector<SimpleGate*>& newGates);
    bool pushVarsDQBFORSplit(SimpleGate* gate, std::vector<SimpleGate*>& newGates);
    bool pushVarsDQBFANDSplit(SimpleGate* gate, std::vector<SimpleGate*>& newGates);
    bool sameEdges(SimpleGate* gate);


    // quantifier localization on QBF, all (will) in hqs_ql_qbf.cpp

    // functions for testing
    void printEdge(SimpleEdge* edge);
    void printVariables(const std::vector<Variable>& variables);
    void printStructure();
    void printGate(SimpleGate* gate);
    void printGateSingle(SimpleGate*);
    void printMacroGate(SimpleGate* gate);
    void printPicture();
};

}  // end namespace hqs

#endif

#include "hqs.ipp"
