#include "hqs.hpp"

#include <aigpp/EdgeRef.hh>
#include <ctime>
#include <iostream>
#include <queue>
#include <vector>


namespace hqs {

void
HenkinSolver::initializeSimpleStructure()
{
    maxVarIndex = 0;
    maxGateID   = 0;
    rootEdge    = nullptr;
    rootGate    = nullptr;
}  // end initializeSimpleStructure



/**
 * \brief makes a copy of the AIG representing the matrix using simpler data structures.
 * \return the SimpleEdge* pointing to the root node of the simple structure.
 * the resulting structure is a circuit consisting of AND gates, OR gates and negations at the edges.
 */

//                                                                      //
//				| rootEdge												//
//				v														//
//				O rootGate g_r (child of g_1 and g_2)					//
//			   / \														//
//			  v   v														//
//		  g_1 O   O g_2 (parents of g_r)								//
//           / \ / \													//
//			v  v v  v													//
//             ...														//
//
// each gate has a gate id.
// if the gate is a variable, then it also holds the HQS variable.
HenkinSolver::SimpleEdge*
HenkinSolver::getSimpleStructure()
{
    // get the rootEdge
    const aigpp::InternalEdgeRef& intEdgeRef  = _matrix.getInternal();
    SimpleEdge*            newRootEdge = new SimpleEdge();
    newRootEdge->isNeg                 = intEdgeRef.isInverted();

    // get rootNode and according rootGate
    aigpp::Node* rootNode = intEdgeRef.node();

    SimpleGate* newRootGate = new SimpleGate();

    maxGateID++;
    newRootGate->id = maxGateID;
    newRootEdge->myGate = newRootGate;

    // needed datastructures
    std::queue<aigpp::Node*>            pendingNodes;
    std::map<aigpp::Node*, SimpleGate*> nodeMap;     // maps aigpp::Node to SimpleGate

    nodeMap[rootNode] = newRootGate;
    pendingNodes.push(rootNode);

    // get all gates and edges
    // each nodes goes in here already has a corresponding gate in nodeMap
    // only process those nodes which do not has parent yet
    while (!pendingNodes.empty()) {
        // get current node and gate
        aigpp::Node* currentNode = pendingNodes.front();
        pendingNodes.pop();

        SimpleGate* currentGate = nodeMap.find(currentNode)->second;

        if (currentGate->var > 0) {
            continue;
        }
        if (currentGate->parent1 != NULL) {
            continue;
        }

        // get parents, create new ones if needed
        const aigpp::Edge& p1Edge = currentNode->parent1();
        const aigpp::Edge& p2Edge = currentNode->parent2();

        aigpp::Node* p1Node = p1Edge.node();
        aigpp::Node* p2Node = p2Edge.node();

        // find parent gates
        SimpleGate* p1Gate;
        SimpleGate* p2Gate;

        // parent1
        if (nodeMap.find(p1Node) != nodeMap.end()) {
            p1Gate = nodeMap.find(p1Node)->second;  // either p1gate already there
        }
        else {
            p1Gate = new SimpleGate();  // or create newly

            maxGateID++;
            p1Gate->id = maxGateID;
            if (p1Node->isVar()) {
                p1Gate->var   = aig2var(p1Node->varIndex());
                if (p1Gate->var > maxVarIndex) {
                    maxVarIndex = p1Gate->var;
                }
                primInputs[p1Gate->var] = p1Gate;
            }
            nodeMap[p1Node] = p1Gate;
        }

        // parent2
        if (nodeMap.find(p2Node) != nodeMap.end()) {
            p2Gate = nodeMap.find(p2Node)->second;
        } else {
            p2Gate = new SimpleGate();

            maxGateID++;
            p2Gate->id = maxGateID;
            if (p2Node->isVar()) {
                p2Gate->var   = aig2var(p2Node->varIndex());
                if (p2Gate->var > maxVarIndex) {
                    maxVarIndex = p2Gate->var;
                }
                primInputs[p2Gate->var] = p2Gate;
            }
            nodeMap[p2Node] = p2Gate;
        }

        // set up parent edges
        SimpleEdge* p1 = new SimpleEdge();
        SimpleEdge* p2 = new SimpleEdge();

        // if gate is an OR, the negations on the parents are to be inverted
        p1->isNeg = p1Edge.isInverted() ^ currentNode->isNAND();
        p2->isNeg = p2Edge.isInverted() ^ currentNode->isNAND();
        currentGate->isOR = currentNode->isNAND();


        p1->myGate = p1Gate;
        p1->origin = currentGate;
        p1Gate->children.insert(p1);

        p2->myGate = p2Gate;
        p2->origin = currentGate;
        p2Gate->children.insert(p2);

        currentGate->parent1 = p1;
        currentGate->parent2 = p2;

        pendingNodes.push(p1Node);
        pendingNodes.push(p2Node);

    }  // end get all gates and edges

    return newRootEdge;

}  // end getSimpleStructure




/**
 * \brief turns structure obtained by getSimpleStructure into NNF.
 */

//        O  O  O  		    		 O   O  O
//		   \!| /! 		 ->  		 |	!| /!
// 			vvv						 v   vv
// 			 O					     O	 O (copy)
// -> copy gets negations
void
HenkinSolver::buildNNF()
{
    std::queue<SimpleGate*> pending;
    seenChildren.clear();
    seenChildren.resize(maxGateID + 1, 0);

    // first check the root
    if (rootEdge->isNeg) {
        rootEdge->isNeg ^= 1;
        rootGate->isOR ^= 1;
        rootGate->parent1->isNeg ^= 1;
        rootGate->parent2->isNeg ^= 1;
    }

    pending.push(rootGate->parent1->myGate);
    pending.push(rootGate->parent2->myGate);

    // now check all other edges

    while (!pending.empty()) {
        SimpleGate* currentGate = pending.front();
        pending.pop();
        ++seenChildren[currentGate->id];

        if (seenChildren[currentGate->id] < currentGate->children.size()) continue;
        if (currentGate->var > 0) continue;

        bool diff = false;
        bool neg  = (*(currentGate->children.begin()))->isNeg;  // mark a difference in negation of the children
                                                                // if true -> need to split

        for (const auto c : currentGate->children) {
            if (neg ^ c->isNeg) {
                diff = true;
                break;
            }
        }

        if (diff) {  // children have different negations -> split
            SimpleGate* newGate = new SimpleGate();
            ++maxGateID;
            newGate->id = maxGateID;

            newGate->isOR = currentGate->isOR ^ 1;

            // currentGate keeps the children which are not negated

            // insert children with negation into newGate
            for (const auto c : currentGate->children) {
                if (c->isNeg) newGate->children.insert(c);
            }

            // delete all children from newGate from currentGate
            // and replace for those children the parent currentGate by newGate
            for (const auto c : newGate->children) {
                currentGate->children.erase(c);
                c->isNeg = false;
                if (c->origin->parent1->myGate == currentGate)
                    c->origin->parent1->myGate = newGate;
                else
                    c->origin->parent2->myGate = newGate;
            }

            // create new edges from newGate to parents of currentGate
            SimpleEdge* newParent1 = new SimpleEdge();
            SimpleEdge* newParent2 = new SimpleEdge();

            newParent1->origin = newGate;
            newParent2->origin = newGate;

            newParent1->myGate = currentGate->parent1->myGate;
            newParent2->myGate = currentGate->parent2->myGate;

            newParent1->isNeg = currentGate->parent1->isNeg ^ 1;
            newParent2->isNeg = currentGate->parent2->isNeg ^ 1;

            newGate->parent1 = newParent1;
            newGate->parent2 = newParent2;

            newParent1->myGate->children.insert(newParent1);
            newParent2->myGate->children.insert(newParent2);

            ++seenChildren[newParent1->myGate->id];
            ++seenChildren[newParent2->myGate->id];
        } 
        else if (neg) {  // all children are negated
            for (const auto c : currentGate->children) c->isNeg = false;

            currentGate->isOR ^= 1;
            currentGate->parent1->isNeg ^= 1;
            currentGate->parent2->isNeg ^= 1;
        }

        pending.push(currentGate->parent1->myGate);
        pending.push(currentGate->parent2->myGate);

    }  // end while

}  // end buildNNF



/**
 * \brief get supports for each gate
 */
void
HenkinSolver::getSupports()
{
	seenParents.clear();
	seenParents.resize(maxGateID+1, false);
	seenGates.clear();
	seenGates.resize(maxGateID+1, false);
	std::queue<SimpleGate*> pending;

	for (const auto p : primInputs) {
		if (p.second->supports.size() == 0) {
			p.second->supports.resize(maxVarIndex+1, false);
			p.second->supports[p.second->var] = true;
		}
		for (const auto c : p.second->children)
			pending.push(c->origin);
	}

	while (!pending.empty()) {
		SimpleGate* currentGate = pending.front();
		pending.pop();

		if (!seenParents[currentGate->id]) {
			seenParents[currentGate->id] = true;
			continue;
		}

		if (seenGates[currentGate->id]) {
			continue;
		}

		currentGate->supports.clear();
		currentGate->supports.resize(maxVarIndex+1, false);

		for (size_t v = 1; v < currentGate->supports.size(); ++v) {
			currentGate->supports[v] =
				currentGate->parent1->myGate->supports[v] || currentGate->parent2->myGate->supports[v];
		}

		seenGates[currentGate->id] = true;

		for (const auto c : currentGate->children) pending.push(c->origin);
	}
}  // end getSupports



/**
 * \brief try to build macrogate with gate as root
 * \return true if gate is root of a macrogate
 */
bool
HenkinSolver::buildMacroGate(SimpleGate* gate)
{
    gate->macroOrigin = gate;

    std::queue<SimpleEdge*> pending;

    pending.push(gate->parent1);
    pending.push(gate->parent2);

    while (!pending.empty()) {
        SimpleEdge* currentEdge = pending.front();
        pending.pop();

        // do not include variable gates, edges with variables, fanout gates into macrogate
        if ((currentEdge->myGate->var > 0)
            || (currentEdge->simplePrefix.size() > 0)
            || (currentEdge->myGate->children.size() > 1)
            || (currentEdge->myGate->isOR != gate->isOR)) {
            gate->macroParents.insert(currentEdge);
        } else {
            currentEdge->myGate->macroOrigin = gate;
            pending.push(currentEdge->myGate->parent1);
            pending.push(currentEdge->myGate->parent2);
        }
    }  // end while

    if (gate->macroParents.size() <= 2) {
        gate->macroOrigin = nullptr;
        gate->macroParents.clear();
        return false;
    } else
        return true;

}  // end buildMacroGate



/**
 * \brief if variable var can be pushed over gate, var must be deleted from all children edges
 */
void
HenkinSolver::deleteVarFromChildren(SimpleGate* gate, Variable var)
{
    // for root gate there is only one edge
    if (gate == rootGate) {
        for (size_t pos = 0; pos < rootEdge->simplePrefix.size(); ++pos) {
            if (rootEdge->simplePrefix[pos] == var) {
                rootEdge->simplePrefix.erase(rootEdge->simplePrefix.begin() + pos);
                return;
            }
        }
    }

    // for normal gate there can be many children
    for (const auto c : gate->children) {
        for (size_t pos = 0; pos < c->simplePrefix.size(); ++pos) {
            if (c->simplePrefix[pos] == var) {
                c->simplePrefix.erase(c->simplePrefix.begin() + pos);
                break;
            }
        }
    }
}  // end deleteVarFromChildren



void
HenkinSolver::clearSimpleStructure()
{
    seenParents.clear();
    seenParents.resize(maxGateID + 1, false);

    std::queue<SimpleEdge*> pending;
    for (auto p : primInputs) {
        for (const auto c : p.second->children) pending.push(c);
        delete p.second;
    }
    primInputs.clear();


    while (!pending.empty()) {
        SimpleEdge* currentEdge = pending.front();
        pending.pop();
        SimpleGate* originGate = currentEdge->origin;

        if (!seenParents[originGate->id]) {
            seenParents[originGate->id] = true;
            continue;
        }

        delete originGate->parent1;
        delete originGate->parent2;

        for (const auto c : originGate->children) pending.push(c);

        delete originGate;
    }

    delete rootEdge;
    rootEdge = nullptr;
    rootGate = nullptr;

    parsWithVar.clear();
    aigs.clear();
    aigvec.clear();

}  // end clearSimpleStructure



HenkinSolver::SimpleGate::SimpleGate() :
    id(0),
    var(0),
    isOR(false),
    parent1(nullptr),
    parent2(nullptr),
    macroOrigin(nullptr)
{}

HenkinSolver::SimpleGate::~SimpleGate()
{}

HenkinSolver::SimpleEdge::SimpleEdge() :
    isNeg(false),
    myGate(nullptr),
    origin(nullptr)
{}

HenkinSolver::SimpleEdge::~SimpleEdge()
{}



// test functions

void
HenkinSolver::printEdge(SimpleEdge* edge) {
	if (edge == rootEdge)
		std::cout << "rootEdge **";
	else std::cout << edge->origin->id ;

	std::cout << " -(" << (edge->isNeg ? 1 : 0) << ")-> " << edge->myGate->id;

	if (edge->simplePrefix.size() > 0) {
		std::cout << " has variables at edge: ";
		printVariables(edge->simplePrefix);
	}
	else std::cout << std::endl;
}

void
HenkinSolver::printVariables(const std::vector<Variable>& variables) {
	for (const auto v : variables) {
		if (_prefix->isUniversal(v)) std::cout << " A" << v;
		else if (_prefix->isExistential(v)) {
			std::cout << " E" << v << "(";
			const auto dep = _dqbf_prefix->getDependencies(v);
			for (const auto d : dep) std::cout << d << " ";
			std::cout << ")";
		}
		else std::cout << " ?" << v;
	}
	std::cout << std::endl;
}



void
HenkinSolver::printStructure()
{
	std::cout << "printing structure" << std::endl;
	if (rootEdge->simplePrefix.size() > 0) {
		std::cout << "	rootEdge with prefix: ";
		printVariables(rootEdge->simplePrefix);
	}
	std::cout << "	rootGate " << (rootGate->isOR ? "OR" : "AND") << " gate " << rootGate->id;
	std::cout << " | supp: ";
	for (size_t i = 1; i <= maxVarIndex; ++i) {
		if (rootGate->supports[i]) std::cout << i << " ";
	}
	std::cout << std::endl;
	std::cout << "		parent 1: " << (rootGate->parent1->myGate->isOR ? "OR" : "AND") << " gate " << rootGate->parent1->myGate->id;
	if (rootGate->parent1->myGate->var > 0) std::cout << ", is var " << rootGate->parent1->myGate->var;
	if (rootGate->parent1->simplePrefix.size() > 0){
		std::cout << ", with prefix ";
		printVariables(rootGate->parent1->simplePrefix);
	}
	else std::cout << std::endl;
	std::cout << "		parent 2: " << (rootGate->parent2->myGate->isOR ? "OR" : "AND") << " gate " << rootGate->parent2->myGate->id;
	if (rootGate->parent2->simplePrefix.size() > 0){
		std::cout << " with prefix ";
		printVariables(rootGate->parent2->simplePrefix);
	}
	else std::cout << std::endl;
	
	std::queue<SimpleGate*> pending;

	pending.push(rootGate->parent1->myGate);
	pending.push(rootGate->parent2->myGate);

	std::vector<unsigned int> seen(maxGateID+1,0);
	while (!pending.empty()) {
		SimpleGate* currentGate = pending.front();
		pending.pop();

		++seen[currentGate->id];

		if (seen[currentGate->id] < currentGate->children.size()) continue;

		printGate(currentGate);
	
		if (currentGate->var == 0) {
			pending.push(currentGate->parent1->myGate);
			pending.push(currentGate->parent2->myGate);
		}
	}
}

void 
HenkinSolver::printGateSingle(SimpleGate* gate)
{
	std::cout << (gate->isOR ? "OR" : "AND") << " gate " << gate->id;
	if (gate->var > 0) std::cout << " (is var " << gate->var << ")";
	std::cout << " | supp:";
	for (Variable i = 1; i <= maxVarIndex; ++i) {
		if (gate->supports[i]) std::cout << " " << i;
	}
}

void
HenkinSolver::printGate(SimpleGate* gate) 
{
	std::cout << "  -------------------------------" << std::endl;
	std::cout << "  | printing Gate " << gate-> id << std::endl;
	if (gate == rootGate) {
		std::cout << "  |	** rootGate ";
		printGateSingle(gate);
		std::cout << std::endl;

		std::cout << "  |	     children:" << std::endl;
		std::cout << "  |		rootEdge";
		if (rootEdge->simplePrefix.size() > 0) {
			std::cout << ", with prefix ";
			printVariables(rootEdge->simplePrefix);
		}
		else std::cout << std::endl;

		std::cout << "  |	     parents:" << std::endl;
		std::cout << "  |		parent1: ";
		printGateSingle(gate->parent1->myGate);
		if (gate->parent1->simplePrefix.size() > 0) {
			std::cout << ", with prefix ";
			printVariables(gate->parent1->simplePrefix);
		}
		else std::cout << std::endl;
		std::cout << "  |		parent2: ";
		printGateSingle(gate->parent2->myGate);
		if (gate->parent2->simplePrefix.size() > 0) {
			std::cout << ", with prefix ";
			printVariables(gate->parent2->simplePrefix);
		}
		else std::cout << std::endl;
	}
	else {
		std::cout << "  |	** ";
		printGateSingle(gate);
		std::cout << std::endl;

		std::cout << "  |	     children:" << std::endl;
		for (const auto c : gate->children) {
			std::cout << "  |		";
			printGateSingle(c->origin);
			if (c->simplePrefix.size() > 0) {
				std::cout << ", with prefix ";
				printVariables(c->simplePrefix);
			}
			else std::cout << std::endl;
		}

		std::cout << "  |	     parents:" << std::endl;
		std::cout << "  |		parent1: ";
		printGateSingle(gate->parent1->myGate);
		if (gate->parent1->simplePrefix.size() > 0) {
			std::cout << ", with prefix ";
			printVariables(gate->parent1->simplePrefix);
		}
		else std::cout << std::endl;
		std::cout << "  |		parent2: ";
		printGateSingle(gate->parent2->myGate);
		if (gate->parent2->simplePrefix.size() > 0) {
			std::cout << ", with prefix ";
			printVariables(gate->parent2->simplePrefix);
		}
		else std::cout << std::endl;

	}
} // end printGate

void
HenkinSolver::printMacroGate(SimpleGate* gate) 
{
	std::cout << "printing macrogate with root " << gate->id << " and macroparents " << std::endl;
	for (const auto p : gate->macroParents) {
		std::cout << "	";
		printGateSingle(p->myGate);
		if (p->simplePrefix.size() > 0) {
			std::cout << ", with prefix ";
			printVariables(p->simplePrefix);
		}
		else std::cout << std::endl;
	}

	std::cout << "   gates in cone:" << std::endl;

	std::queue<SimpleGate*> pending;
	if (gate->var == 0) pending.push(gate);

	std::vector<bool> seen(maxGateID+1,false);

	while (!pending.empty()) {
		SimpleGate* currentGate = pending.front();
		pending.pop();

		if (seen[currentGate->id]) continue;

		seen[currentGate->id] = true;

		printGate(currentGate);

		if (currentGate->parent1->myGate->var == 0) {
			if (gate->macroParents.find(currentGate->parent1) == gate->macroParents.end()) {
				pending.push(currentGate->parent1->myGate);
			}
			//else std::cout << "	parent1 is in macroparents" << std::endl;
		}
		//else std::cout << "	parent1 is var " << currentGate->parent1->myGate->var << std::endl;
		if (currentGate->parent2->myGate->var == 0) {
			if (gate->macroParents.find(currentGate->parent2) == gate->macroParents.end()) {
				pending.push(currentGate->parent2->myGate);
			}
			//else std::cout << "	parent2 is in macroparents" << std::endl;
		}
		//else std::cout << "	parent2 is var " << currentGate->parent2->myGate->var << std::endl;
	}
}


void
HenkinSolver::printPicture()
{
	std::cout << "printPicture:" << std::endl;
	std::vector<int> seen(maxGateID+1, false);

	std::vector<SimpleGate*> level;
	std::vector<SimpleGate*> next;
	level.push_back(rootGate);

	while (level.size() > 0) {
		for (const auto l : level) {
			++seen[l->id];
			if (static_cast<unsigned int>(seen[l->id]) < l->children.size()) continue;

			std::cout << l->id;
			if (l->var > 0) std::cout << "(var " << l->var << ")";
			std::cout << " ";

			if (l->var == 0) {
				next.push_back(l->parent1->myGate);
				next.push_back(l->parent2->myGate);
			}
		
		}
		std::cout << std::endl;
		std::cout << std::endl;

		level.clear();
		level = next;
		next.clear();
	}
}







} // end namespace
