#include "hqs.hpp"

#include <aigpp/EdgeRef.hh>
#include <iostream>
#include <limits>
#include <queue>
#include <vector>
#include <chrono>

namespace hqs {


/**
 * \brief General procedure for quantifier localization on DQBF
 *
 * \return TRUE/FALSE if the truth value of the formula could be determined
 *          during quantification, UNKNOWN otherwise.
 */
TruthValue
HenkinSolver::earlyQuantifyDQBF() 
{
	// simple case: if matrix is constant or only one variable, then we do not do anything
	if (_matrix.nodeCount() < 2) {
		return TruthValue::UNKNOWN;
	}

	if (cleanUpUnnecessaryVars()) {
		return TruthValue::UNKNOWN;
	}

	initializeSimpleStructure();

	rootEdge = getSimpleStructure();

	rootGate = rootEdge->myGate;

	buildNNF();

	getSupports();

	// push variables into the formula
    const bool pushed = localizeQuantifiersDQBF();

	if (!pushed) {
        clearSimpleStructure();
        return TruthValue::UNKNOWN;
    }

    getSupports();

	std::vector<std::set<SimpleEdge*>> varAtEdges(maxVarIndex + 1, std::set<SimpleEdge*>());

    // connect prefixes topologically and also count how often a variable occur in a prefix
    findPrefixTopologyAndOccurrences(rootEdge, varAtEdges);

	// delete all dependencies of Ey on Ax if x does not occur on any path from an occurrence of y to the root
    deleteDeps(varAtEdges);

    // if no local eliminations were possible, then do not use new EdgeRef
	cntQuantifications = 0;

	// while building AIG, quantify whenever possible
	aigpp::EdgeRef newEdgeRef;
    buildEarlyQuantifiedDQBF(newEdgeRef, varAtEdges);

    clearSimpleStructure();

	if (cntQuantifications == 0) {
        return TruthValue::UNKNOWN;
	}

    if (aigpp::functionallyFalse(newEdgeRef)) {
        return TruthValue::FALSE;
    }
    else if (aigpp::functionallyTrue(newEdgeRef)) {
        return TruthValue::TRUE;
    }
    else {
        _matrix = newEdgeRef;
        return TruthValue::UNKNOWN;
    }
} // end earlyQuantifyDQBF




/**
 * \brief remove all variables which are deleted already
 * eliminate all universal variables on which no existential variable depends
 * eliminate all existential variables which depend on all universal variables
 * \return true if only one kind of quantifier is left
 */
bool
HenkinSolver::cleanUpUnnecessaryVars()
{
	std::vector<Variable> toRemove;
	_aig_buffer.clear();
	bool again = false;

	if (_dqbf_prefix->numUVars() == 0 || _dqbf_prefix->numEVars() == 0) return true;

	while (true) {
		const auto univVars = _dqbf_prefix->getUnivVars();

		for (const auto u : univVars) {
			if (_dqbf_prefix->getDependencies(u).size() == 0) {
				_aig_buffer.push_back(var2aig(u));
				toRemove.push_back(u);
			}
		}

		if (_aig_buffer.size() > 0) {
			_manager.bddQuantifyInplace(_matrix,_aig_buffer,false,false,true);

			_manager.makeFRAIG(10);

			for (const auto u : toRemove) removeVar(u);
			toRemove.clear();
			_aig_buffer.clear();
			again = true;
		}
		
		const auto existVars = _dqbf_prefix->getExistVars();

		for (const auto e : existVars) {
			if (_dqbf_prefix->getDependencies(e).size() == _dqbf_prefix->getUnivVars().size()) {
				_aig_buffer.push_back(var2aig(e));
				toRemove.push_back(e);
			}
		}

		if (_aig_buffer.size() > 0) {

			_manager.bddQuantifyInplace(_matrix,_aig_buffer,true,false,true);
			_manager.makeFRAIG(10);

			for (const auto e : toRemove) removeVar(e);
			toRemove.clear();
			_aig_buffer.clear();
			again = true;
		}
		else again = false;

		if (!again) break;
		
	} // end while

	if (_prefix->numUVars() == 0 || _prefix->numEVars() == 0) return true;
	else return false;

} // end cleanUpUnecessaryVars




/**
 * \brief performs quantifier localization = 
 * try to push variables as far as possible into the formula
 */
bool
HenkinSolver::localizeQuantifiersDQBF()
{
    const std::vector<Variable>& exVars   = _prefix->getExistVars();
    const std::vector<Variable>& univVars = _prefix->getUnivVars();

    // insert variables to the prefix at rootEdge
    // for a DQBF we keep the two blocks Ax_1,...,x_n Ey_1,...,y_m
    for (const auto u : univVars) {
    	if (u > maxVarIndex) { // variable was not in cone -> delete
    		removeVar(u);
    		continue;
    	}
    	if (!rootGate->supports[u]) {
    		removeVar(u); // variable is not in support, thus also not in formula -> delete

    		continue;
    	}
            
        rootEdge->simplePrefix.push_back(u);
    }
    for (const auto e : exVars) {
    	if (e > maxVarIndex) {
    		removeVar(e);
    		continue;
    	}
        if (!rootGate->supports[e]) {
            removeVar(e);             
            continue;
        }
            
        rootEdge->simplePrefix.push_back(e);
    }

    bool pushed = false;
    std::queue<SimpleEdge*> pending;
	std::vector<SimpleGate*> newGates;
	std::vector<Variable> commonVars;

    commonVars = rootEdge->simplePrefix;

	// first try root
	// if no variables pushed over root, no need to run through the graph
	if (!buildMacroGate(rootGate)) { // gate is not origin of a macroGate
		if (pushVarsDQBF(rootEdge, commonVars)) {
			pushed = true;
		}
		else {
			return false;
		}

		if (rootGate->parent1->myGate->var == 0) {
			pending.push(rootGate->parent1);
		}
		if (rootGate->parent2->myGate->var == 0) {
			pending.push(rootGate->parent2);
		}
	} 
	else { // gate is origin of a macrogate
		if (pushVarsMacroDQBF(rootEdge, commonVars)) {
			pushed = true;
		}
		else {
			return false;
		}

		for (const auto p : rootGate->macroParents) {
			if (p->myGate->var == 0) pending.push(p);
		}
	}

	
    seenChildren.clear();
    seenChildren.resize(maxGateID + 1, 0);


	// check from root to inputs
	// for each gate whether it is the root of a macrogate and 
	// whether a variable can be pushed over this gate
    while (!pending.empty()) {
        SimpleEdge* currentEdge = pending.front();
        pending.pop();
        SimpleGate* currentGate = currentEdge->myGate;

        ++seenChildren[currentGate->id];

        // did not see all children yet
        if (currentGate->children.size() > seenChildren[currentGate->id]) {
        	continue;
        }
	
	
		if (!buildMacroGate(currentGate)) { // gate is not origin of a macroGate
			if (currentGate->children.size() == 1) {
				if (currentEdge->simplePrefix.size() > 0) {
					commonVars = currentEdge->simplePrefix;
					if (pushVarsDQBF(currentEdge, commonVars)) {
						pushed = true;
					}
				}
			}
			else if (pushVarsDQBF2(currentEdge, newGates)) {
				pushed = true;
			}

			if (currentGate->parent1->myGate->var == 0) {
				pending.push(currentGate->parent1);
			}
			if (currentGate->parent2->myGate->var == 0) {
				pending.push(currentGate->parent2);
			}
			for (const auto g : newGates) {
				if (g->parent1->myGate->var == 0) {
					pending.push(g->parent1);
				}
				if (g->parent2->myGate->var == 0) {
					pending.push(g->parent2);
				}
			}
		} 
		else { // gate is origin of a macrogate
			if (currentGate->children.size() == 1) {
				if (currentEdge->simplePrefix.size() > 0) {
					commonVars = currentEdge->simplePrefix;
					if (pushVarsMacroDQBF(currentEdge, commonVars)) {
						pushed = true;
					}
				}
			}
			else if (pushVarsMacroDQBF2(currentEdge, newGates)) {
				pushed = true;
			}

			for (const auto p : currentGate->macroParents) {
				if (p->myGate->var == 0) pending.push(p);
			}
			for (const auto g : newGates) {
				for (const auto p : g->macroParents) {
					if (p->myGate->var == 0) pending.push(p);
				}
			}
		}
		newGates.clear();
    }  // end while
    return pushed;

}  // end localizeQuantifiersDQBF


/**
 * \brief connect prefixes topologically,
 * count how often a var occurs in a prefix,
 * fill seenUnivVars for edges with prefix
 */
void
HenkinSolver::findPrefixTopologyAndOccurrences(SimpleEdge* edge, std::vector<std::set<SimpleEdge*>>& varAtEdges)
{
    if (edge->myGate->var > 0) return;

    // need two flags to find lowest prefixes?!?!
    seenChildren.clear();
    seenChildren.resize(maxGateID + 1, 0);
    seenGates.clear();
    seenGates.resize(maxGateID + 1, false);
    seenEdges.clear();

    // starting from inputted edge, find all topmost prefixes, which is not in the scope of other variables
    std::queue<SimpleEdge*> pendingEdges;
    // abusing this queue to find top most prefixes. Is actually prepared for the next section
    std::queue<SimpleEdge*> pendingSubEdges;

    pendingSubEdges.push(edge);

    // find topMost edges with prefix
    while (!pendingSubEdges.empty()) {
        SimpleEdge* currentEdge = pendingSubEdges.front();
        pendingSubEdges.pop();
        
        if (currentEdge->myGate->var > 0) continue;

        if (currentEdge->simplePrefix.size() > 0) {
            pendingEdges.push(currentEdge);
            seenGates[currentEdge->myGate->id] = true;
        } 
        else {
            if (seenGates[currentEdge->myGate->id]) continue;

            seenChildren[currentEdge->myGate->id]++;
            if (seenChildren[currentEdge->myGate->id] >= currentEdge->myGate->children.size()) {
                pendingSubEdges.push(currentEdge->myGate->parent1);
                pendingSubEdges.push(currentEdge->myGate->parent2);
            }
        }
    }  // end while

    // go through all edges and connect them topologically
    while (!pendingEdges.empty()) {
        SimpleEdge* currentEdge = pendingEdges.front();
        pendingEdges.pop();

        // if (seenEdges.find(currentEdge) != seenEdges.end()) continue;

        seenEdges.insert(currentEdge);

        // first count occurrence of the vars in this prefix, and store, at which edges a var occur
        // vars at edges pointing to var gates are counted in the end, since these edges are never counted
        for (int i = currentEdge->simplePrefix.size() - 1; i >= 0; --i) {
            if (_prefix->isUniversal(currentEdge->simplePrefix[i]))
                currentEdge->seenUnivVars.insert(currentEdge->simplePrefix[i]);

            varAtEdges[currentEdge->simplePrefix[i]].insert(currentEdge);
        }

        if (currentEdge->myGate->var > 0) continue;

        seenGates.clear();
        seenGates.resize(maxGateID + 1, false);

        pendingSubEdges.push(currentEdge->myGate->parent1);
        pendingSubEdges.push(currentEdge->myGate->parent2);

        // for currentEdge, find those prefixes which are directly below it
        while (!pendingSubEdges.empty()) {
            SimpleEdge* currentSubEdge = pendingSubEdges.front();
            pendingSubEdges.pop();

            if (currentSubEdge->simplePrefix.size() > 0) {
                currentSubEdge->prefixChildren.insert(currentEdge);
                currentEdge->prefixParents.insert(currentSubEdge);

                currentSubEdge->seenUnivVars = currentEdge->seenUnivVars;
                if ((currentSubEdge->myGate->var == 0) && (seenEdges.find(currentSubEdge) == seenEdges.end())) {
                    pendingEdges.push(currentSubEdge);
                }
            } else {
                if (!seenGates[currentSubEdge->myGate->id]) {
                    if (currentSubEdge->myGate->var == 0) {
                        pendingSubEdges.push(currentSubEdge->myGate->parent1);
                        pendingSubEdges.push(currentSubEdge->myGate->parent2);
                    }
                }
            }

            seenGates[currentSubEdge->myGate->id] = true;
        }  // end while iterating all subedges to find next edge with var

    }  // end while iterating all Edges

    // count var occurrences at edges pointing to var gates
    // not that nice -> rewrite
    for (const auto p : primInputs) {
        for (const auto c : p.second->children) {
            if (c->simplePrefix.size() > 0) {
                for (int i = c->simplePrefix.size() - 1; i >= 0; --i) {
                    if (_prefix->isUniversal(c->simplePrefix[i])) c->seenUnivVars.insert(c->simplePrefix[i]);

                    varAtEdges[c->simplePrefix[i]].insert(c);
                }

                // lowest.push_back(c);
            }
        }
    }
}  // end findPrefixTopologyAndOccurrences



/**
 * if for an existential variable y it holds that a universal variable x from y's dependency set
 * does not occur on any path from any occurrence of y to the root, then we can delete the dependency
 */
void
HenkinSolver::deleteDeps(std::vector<std::set<SimpleEdge*>>& varAtEdges)
{
    for (size_t i = 1; i < varAtEdges.size(); ++i) {
        if (_prefix->isExistential(i)) {
            const std::set<Variable>& deps = _dqbf_prefix->getDependencies(i);

            for (const auto d : deps) {
                bool dependent = false;
                for (const auto e : varAtEdges[i]) {
                    if (e->seenUnivVars.find(d) != e->seenUnivVars.end()) {
                        dependent = true;
                        break;
                    }
                }

                if (dependent) break;
                _dqbf_prefix->removeDependency(i, d);
            }
        }
    }
}  // end deleteDeps


/**
 * \brief build EdgeRef according to new structure,
 * quantify localized variables whereever possible,
 * for each gate: call buildAIG
 */
void
HenkinSolver::buildEarlyQuantifiedDQBF(aigpp::EdgeRef& newRootEdgeRef, std::vector<std::set<SimpleEdge*>>& varAtEdges)
{
    aigvec.resize(maxGateID + 1);
    std::queue<SimpleGate*> pending;

    for (const auto p : primInputs) {
        aigvec[p.second->id] = var2aig(p.first);
        for (const auto c : p.second->children) pending.push(c->origin);
    }

    seenParents.clear();
    seenParents.resize(maxGateID + 1, false);

    // build aig for each gate from inputs to outputs.
    while (!pending.empty()) {
        SimpleGate* currentGate = pending.front();
        pending.pop();

        if (!seenParents[currentGate->id]) {
            seenParents[currentGate->id] = true;
            continue;
        }

        buildAIG(currentGate, varAtEdges);

        if (currentGate == rootGate) {
            newRootEdgeRef = aigvec[currentGate->id];
            newRootEdgeRef = newRootEdgeRef.notIf(rootEdge->isNeg); // actually this cannot happen, except the root gate is also a var gate

            if (rootEdge->simplePrefix.size() > 0) {
                quantify(rootEdge, newRootEdgeRef, varAtEdges);
            }

            return;
        } else {
            for (const auto c : currentGate->children) pending.push(c->origin);
        }
    }  // end while

}  // end buildEarlyQuantified




/**
 * push variables written at this edge over its gate to both parents.
 * \return true, if have pushed something
 */
bool
HenkinSolver::pushVarsDQBF(SimpleEdge* edge, std::vector<Variable>& commonVars)
{
	// can only push variables which comes along all children edges
    //if (commonVars.size() == 0) return false;

    SimpleGate* gate = edge->myGate;

    bool pushed = false;

    std::set<Variable> stuck;

    if (!gate->isOR)  // and gate
    {
        for (int i = commonVars.size() - 1; i >= 0; --i)  // first check existential vars and then univ vars
        {
            Variable var = commonVars[i];
			
			// can only push an existential variable over an AND gate if it is just in one parent.
            if (_prefix->isExistential(var)) {
                if (gate->parent1->myGate->supports[var] && gate->parent2->myGate->supports[var]) {
                    stuck.insert(var);
                } else if (gate->parent1->myGate->supports[var]) {
                    gate->parent1->simplePrefix.push_back(var);
                    deleteVarFromChildren(gate, var);
                    pushed = true;
                } else if (gate->parent2->myGate->supports[var]) {
                    gate->parent2->simplePrefix.push_back(var);
                    deleteVarFromChildren(gate, var);
                    pushed = true;
                }
            } 
            // can only push universal variable x if no ex var with x in dependency set is still at the edge 
            else if (_prefix->isUniversal(var)) {
                const std::set<Variable>& deps = _dqbf_prefix->getDependencies(var);

                bool stuckHere = false;
                for (const auto d : deps) {
                    if (stuck.find(d) != stuck.end()) {
                        stuckHere = true;
                        break;
                    }
                }

                if (!stuckHere) {
                    if (gate->parent1->myGate->supports[var])
                        gate->parent1->simplePrefix.insert(gate->parent1->simplePrefix.begin(), var);
                    if (gate->parent2->myGate->supports[var])
                        gate->parent2->simplePrefix.insert(gate->parent2->simplePrefix.begin(), var);

                    deleteVarFromChildren(gate, var);
                    pushed = true;
                }
            }
        }
    }     // end and gate
    else  // or gate
    {
        for (int i = commonVars.size() - 1; i >= 0; --i) {
            Variable var = commonVars[i];

            if (_prefix->isExistential(var))  // distribute existential vars if possible
            {
				if (gate->parent1->myGate->supports[var]) {
					if (gate->parent2->myGate->supports[var]) {
						std::set<SimpleEdge*> parents;
						parents.insert(gate->parent2);
						parents.insert(gate->parent1);
						if (!isPushableEx(parents,var,gate)) {
							stuck.insert(var);
							continue;
						}
						else {
                    		gate->parent1->simplePrefix.push_back(var);
                    		gate->parent2->simplePrefix.push_back(var);
						}
					}
					else
                    	gate->parent1->simplePrefix.push_back(var);
				}
				else if (gate->parent2->myGate->supports[var])
                    gate->parent2->simplePrefix.push_back(var);

                deleteVarFromChildren(gate, var);
                pushed = true;

            } // end ex vars 
            else  // only push universal vars which occur only in one parent and in the other parent
            {
            	if (gate->parent1->myGate->supports[var]) {
            		if (gate->parent2->myGate->supports[var]) {
            			// cannot push, var in both parents
            			continue;
            		} 
            		else {
            			const auto deps = _dqbf_prefix->getDependencies(var);

            			bool stuckhere = false;
            			for (const auto d : deps) {
            				if (stuck.find(d) != stuck.end()) {
            					stuckhere = true;
            					break;
            				}
            				if (gate->parent2->myGate->supports[d]) {
            					stuckhere = true;
            					break;
            				}
            			}

            			if (stuckhere) continue;
                    
						gate->parent1->simplePrefix.insert(gate->parent1->simplePrefix.begin(), var);
						deleteVarFromChildren(gate, var);
						pushed = true;
					}

            	}
            	else if (gate->parent2->myGate->supports[var]) {
            		const auto deps = _dqbf_prefix->getDependencies(var);

					bool stuckhere = false;
					for (const auto d : deps) {
						if (stuck.find(d) != stuck.end()) {
							stuckhere = true;
							break;
						}
						if (gate->parent1->myGate->supports[d]) {
							stuckhere = true;
							break;
						}
					}

					if (stuckhere) continue;
				
					gate->parent2->simplePrefix.insert(gate->parent2->simplePrefix.begin(), var);
					deleteVarFromChildren(gate, var);
					pushed = true;
				}	
            } // end univ var
        }  // end for all common variables
    }      // end or gate

    return pushed;

}  // end pushVars





// return bool if pushed any variable over gate
bool
HenkinSolver::pushVarsMacroDQBF(SimpleEdge* edge, std::vector<Variable>& commonVars)
{
    SimpleGate* gate = edge->myGate;

    if (!gate->isOR) {  // and gate
        return pushVarsMacroAnd(gate, commonVars);
    } 
    else  { // or gate
        return pushVarsMacroOr(gate, commonVars);
    }

}  // end pushVarsMacro




/*
 * \brief computes variables which are common in all incoming edges
 * univ var x for which an ex var y with x in Dy is not common in all edges cannot be pushed too
 */
void
HenkinSolver::computeCommonVarsDQBF(SimpleEdge* edge, std::vector<Variable>& commonVars)
{
    std::vector<unsigned int> cnt(maxVarIndex + 1, 0);

    for (const auto c : edge->myGate->children) {
    	// if there is at least one child without any prefix, then there can not be any vars in common
        if (c->simplePrefix.size() == 0)  return;

        for (const auto v : c->simplePrefix) cnt[v]++;
    }

    std::vector<bool> stuck(maxVarIndex + 1, false);
    std::vector<bool> common(maxVarIndex + 1, false);

	for (const auto v : edge->simplePrefix) {
		if (cnt[v] < edge->myGate->children.size()) stuck[v] = true;
		else common[v] = true;
	}

	for (const auto v : edge->simplePrefix) {
		if (!common[v]) continue;

		if (_prefix->isExistential(v)) commonVars.push_back(v);
		else if (_prefix->isUniversal(v)) {
			const auto deps = _dqbf_prefix->getDependencies(v);
			bool pushable = true;

			// if a stucked variable depends on v, v cannot be pushed
			for (const auto d : deps) {
				if (stuck[d]) {
					pushable = false;
					break;
				}
			}

			if (pushable) commonVars.insert(commonVars.begin(), v);
		}
	}
}  // end computeCommonVarsDQBF


/**
 * \brief checks for an existential variable, whether it can be pushed over a disjunction
 */
bool
HenkinSolver::isPushableEx(const std::set<SimpleEdge*>& parents, Variable exVar, SimpleGate* gate) 
{
	const auto deps = _dqbf_prefix->getDependencies(exVar);
	std::map<Variable, SimpleEdge*> varParentsMap;

	// variables which may not occ outside of subformula (not even in dep set)
	// univ var x which is in branch of subformula, but not in D_exVar
	// univ var x which is in dep set of branch of subformula, but not in D_exVar
	std::vector<bool> disjunct(maxVarIndex+1, false); 

	// check whether the parents have disjunct supports regarding univ vars not in dep set of exVar 
	// and univ vars that are in dep set of other ex vars, but not in dep set of exVar.
	// also existential variables with x in support which is not in D_exVar will be collected here
	for (const auto p : parents) {
		std::vector<Variable> currentDisjunct;

		// collect all universal variables which are not in dep set of exVar
		for (size_t i = 1; i < p->myGate->supports.size(); ++i) {
			if (!p->myGate->supports[i]) continue;

			if (_prefix->isUniversal(i)) {
				if (deps.find(i) == deps.end()) { // univ var not in dep set of exVar
					if (disjunct[i]) {
						return false;
					}
					else {
						currentDisjunct.push_back(i);
					}
				}
			}
			else {
				if (disjunct[i]) {
					return false;
				}
				const auto otherDeps = _dqbf_prefix->getDependencies(i);

				for (const auto d : otherDeps) {
					if (deps.find(d) == deps.end()) {
						if (disjunct[d]) {
							return false;
						}
						else {
							currentDisjunct.push_back(i);
							currentDisjunct.push_back(d);
						}
					}
				}
			}
		}


		for (const auto d : currentDisjunct) {
			disjunct[d] = true;
			varParentsMap[d] = p;
		}

	} // end checking whether all parents have disjunct univ vars
	if (gate == rootGate) {
		return true;
	}

	// There can only be one branch that do not fulfill the following conditions.
	// the universal variables from the scope and not in dep set Dy cannot occur anywhere else. 
	// Not even in a dep set.

	// here we already know, that the universal variables in the parent branches are disjunct.

	// check all existential variables which occur outside the formula, 
	// whether they dep on a univ var they shouldn't
	for (Variable i = 1; i <= maxVarIndex; ++i) {
		if (disjunct[i]) continue; // check these variables later

		if (_prefix->isExistential(i) && i != exVar) {
			const auto otherDeps = _dqbf_prefix->getDependencies(i);

			for (const auto d : otherDeps) {
				if (disjunct[d]) {
					return false;
				}
			}
		}
	}

	std::vector<bool> seen(maxGateID+1, false);
	std::queue<SimpleEdge*> pending;
	std::vector<SimpleEdge*> startingPoints;
	std::set<unsigned int> cone;
	
	for (const auto p : parents) collectCone(p->myGate, cone);

	SimpleEdge* notSatConds = NULL;

	// check whether for more than one parent there occurs a variable from disjunct outside the subformula 

	for (size_t var = 1; var < disjunct.size(); ++var) {
		if (!disjunct[var]) continue;

		for (auto c : gate->children) pending.push(c);


		// find edges where var is quantified
		// it is sufficient to start the search from there
		while (!pending.empty()) {
			SimpleEdge* currentEdge = pending.front();
			pending.pop();

			bool foundVar = false;

			if (currentEdge->simplePrefix.size() > 0) {
				for (const auto v : currentEdge->simplePrefix) {
					if (v == var) {
						startingPoints.push_back(currentEdge);
						foundVar = true;
						break;
					}
				}
			}

			if (foundVar) continue;

			if (currentEdge == rootEdge) continue;

			if (currentEdge->origin == rootGate) {
				pending.push(rootEdge);
			}
			else {
				for (auto c : currentEdge->origin->children) pending.push(c);
			}
		} // end finding starting points

		while (!pending.empty()) pending.pop();

		// begin the search from starting points to leaves

		seen.clear();
    	seen.resize(maxGateID + 1, false);

		// if we did not find a variable which must be distinct and occurs above gate then we can assume
		// gate is either not in its scope or the quantification occurs below gate
		if (startingPoints.size() == 0) return true;


		for (auto s : startingPoints) pending.push(s);

		startingPoints.clear();


		while (!pending.empty()) {
			SimpleEdge* currentEdge = pending.front();
			pending.pop();

			if (seen[currentEdge->myGate->id]) continue;
			seen[currentEdge->myGate->id] = true;

			if (parents.find(currentEdge) != parents.end()) continue;

			if (cone.find(currentEdge->myGate->id) != cone.end()) continue;

			if (currentEdge->myGate->var == var) {
				if (notSatConds == NULL) notSatConds = varParentsMap.find(var)->second;
				else if (varParentsMap.find(var)->second != notSatConds) {
					return false;
				}
				
				continue;
			}

			if (currentEdge->myGate->var > 0) continue;

			pending.push(currentEdge->myGate->parent1);
			pending.push(currentEdge->myGate->parent2);
		}
	} // end for all vars which must be disjunct

	return true;

} // end isPushableEx



/**
 * \brief getting cone of a gate
 */
void
HenkinSolver::collectCone(SimpleGate* gate, std::set<unsigned int>& cone) 
{
	std::vector<bool> seen(maxGateID+1, false);

	std::queue<SimpleGate*> pending;
	pending.push(gate->parent1->myGate);
	pending.push(gate->parent2->myGate);

	while (!pending.empty()) {
		SimpleGate* currentGate = pending.front();
		pending.pop();

		if (seen[currentGate->id]) continue;
		seen[currentGate->id] = true;

		cone.insert(currentGate->id);
		if (currentGate->var == 0) {
			pending.push(currentGate->parent1->myGate);
			pending.push(currentGate->parent2->myGate);
		}
	}
} // end collectCone



/**
 * \brief try to push variabels over an AND gate
 * \return true of at least one variable has been pushed
 */
bool
HenkinSolver::pushVarsMacroAnd(SimpleGate* gate, std::vector<Variable>& commonVars)
{
    size_t origSize = commonVars.size();

    getParsWithVar(gate, commonVars);

    while (commonVars.size() > 0) {
        // if (needNew) getParsWithVar(gate,commonVars);

        // smallestVar is taken out of commonVars
        // make decision wrt dependencies
        // an univ var from the dep set of an ex var cannot
        // be pushed before the ex var

        unsigned int nextVar = findSmallestVarAND(commonVars);

		// if next var is existential, we check here, whether all parents have var in support
        if (_prefix->isUniversal(nextVar)) {
            distributeVarsMacro(gate, nextVar);
            deleteVarFromChildren(gate, nextVar);
            parsWithVar.erase(nextVar);
        } else if (_prefix->isExistential(nextVar)) {
            if (parsWithVar.find(nextVar)->second.size() == gate->macroParents.size()) {
                //commonVars.push_back(nextVar);
                break;
            }

            if (parsWithVar.find(nextVar)->second.size() > 1) {
                restructureMacroParents(gate, nextVar);
            }

            (*(parsWithVar.find(nextVar)->second).begin())->simplePrefix.push_back(nextVar);
            deleteVarFromChildren(gate, nextVar);
            parsWithVar.erase(nextVar);
        }

        for (size_t i = 0; i < commonVars.size(); ++i) {
        	if (commonVars[i] == nextVar) {
        		commonVars.erase(commonVars.begin() + i);
        		break;
        	}
        }
    }  // end iterating commonVars

    if (origSize > commonVars.size())
        return true;
    else
        return false;

}  // end pushVarsMacroAnd


/** 
 * \brief try to push variables over an OR gate
 * \return true if at least one variable has been pushed
 */
bool
HenkinSolver::pushVarsMacroOr(SimpleGate* gate, std::vector<Variable>& commonVars)
{
    size_t origSize = commonVars.size();

    // first try to distribute ex vars to macroparents
    for (int i = commonVars.size() - 1; i >= 0; --i) {
    	// if found a universal variable, stop distributing
        if (_prefix->isUniversal(commonVars[i])) {
            break;
        }
		// distribute ex vars if possible
		else {
			std::set<SimpleEdge*> parentsWithVar;
			for (const auto p : gate->macroParents) {
				if (p->myGate->supports[commonVars[i]]) {
					parentsWithVar.insert(p);
				}
			}
			if (parentsWithVar.size() > 1) {
				if (!isPushableEx(parentsWithVar, commonVars[i], gate)) {
					continue;
				}
			}

			// distribute existential vars
			distributeVarsMacro(gate, commonVars[i]);
			deleteVarFromChildren(gate, commonVars[i]);
			commonVars.erase(commonVars.begin() + i);
		}
    } // end for all ex vars which are pushable

    getParsWithVar(gate, commonVars);

    // push vars if possible
    while (commonVars.size() > 0) {
		
		// smallestVar is taken out of commonVars
		// commonVars is short one element after this
        unsigned int nextVar = findSmallestVarOR(gate, commonVars);

        if (nextVar == 0) {
            break;
        }

        if (parsWithVar.find(nextVar)->second.size() == gate->macroParents.size()) {
            break;
        }

        // in all other parents, there cannot be an ex var that dep on this univ var.

        // univ var can be pushed
        if (_prefix->isUniversal(nextVar)) restructureMacroParentsUnivOR(gate, nextVar);
        else restructureMacroParents(gate, nextVar);

		// here there is only one parent left
		if (_prefix->isUniversal(nextVar)) {
        	(*(parsWithVar.find(nextVar)->second).begin())
            	->simplePrefix.insert((*(parsWithVar.find(nextVar)->second).begin())->simplePrefix.begin(),
                                  nextVar);          }
        else {
        	(*(parsWithVar.find(nextVar)->second).begin())->simplePrefix.push_back(nextVar);
        }

        deleteVarFromChildren(gate, nextVar);

        parsWithVar.erase(nextVar);

    }  // end loop until no more commonVars or no vars pushable

    if (origSize > commonVars.size())
        return true;
    else
        return false;

}  // end pushVarsMacroOr



/**
 * \brief store for a variable all macroParents (=edges) where this var is in support
 */
void
HenkinSolver::getParsWithVar(SimpleGate* gate, std::vector<Variable>& vars)
{
	parsWithVar.clear();

	for (const auto v : vars) {
		std::set<SimpleEdge*> pars;

		for (const auto p : gate->macroParents) {
			if (p->myGate->supports[v]) pars.insert(p);
		}
		parsWithVar[v] = pars;
	}
}  // end gerParsWithVar


/**
 * find variable whose pushing summarizes the fewest macroParents
 * TODO: is this still correctly applied?
 */
unsigned int
HenkinSolver::findSmallestVarAND(std::vector<Variable>& vars)
{
	unsigned int smallestVar = 0;
	unsigned int fewestParents = 0;

	for (const auto v : vars) {
		if (_prefix->isUniversal(v)) {
			if (!isPushableUniv(v)) continue;
		}

		if ((smallestVar == 0) || (fewestParents > parsWithVar.find(v)->second.size())) {

			smallestVar = v;
			fewestParents = parsWithVar.find(v)->second.size();
		}
	}

	return smallestVar;
	
}  // end findSmallestVarAND




/*
 * \brief distribute variable to all macroParents, regardless whether gate is OR or AND
 */
void
HenkinSolver::distributeVarsMacro(SimpleGate* gate, Variable var)
{
    for (const auto m : gate->macroParents) {
        if (m->myGate->supports[var]) {
            if (_prefix->isUniversal(var))
                m->simplePrefix.insert(m->simplePrefix.begin(), var);
            else if (_prefix->isExistential(var))
                m->simplePrefix.push_back(var);
        }
    }
}  // end distributeVarsMacro



/**
 * \brief restructure the macroGate such that there is only one parent for this var to enable pushing
 */
void
HenkinSolver::restructureMacroParents(SimpleGate* gate, const unsigned int var)
{
    std::vector<SimpleGate*> orderedGates;

    std::set<SimpleEdge*>& pars = parsWithVar.find(var)->second;
    while (pars.size() > 1) {
        findTopoOrder(gate, orderedGates);

        SimpleEdge* exchange1 = nullptr;
        SimpleEdge* exchange2 = nullptr;

        for (int i = orderedGates.size() - 1; i >= 0; --i) {
            // found a gate where both parents are parents with var
            if ((pars.find(orderedGates[i]->parent1) != pars.end())
                && (pars.find(orderedGates[i]->parent2) != pars.end())) {
                exchange1 = orderedGates[i]->parent1;
                exchange2 = orderedGates[i]->parent2;
                break;
            }
            // only found one parent with par with var
            else if (pars.find(orderedGates[i]->parent1) != pars.end()) {
                if (exchange1 == nullptr) {
                    exchange1 = orderedGates[i]->parent1;
                } else {
                    exchange2 = orderedGates[i]->parent1;
                    break;
                }
            } else if (pars.find(orderedGates[i]->parent2) != pars.end()) {
                if (exchange1 == nullptr) {
                    exchange1 = orderedGates[i]->parent2;
                } else {
                    exchange2 = orderedGates[i]->parent2;
                    break;
                }
            }

        }  // end for current orderedGates

        if ((exchange1 != nullptr) && (exchange2 != nullptr)) {
            if (exchange1->origin != exchange2->origin) {
                swapParents(exchange1, exchange2);
            }
        } 
    
        // replace summarized parents by singleparent
        gate->macroParents.erase(exchange1);
        gate->macroParents.erase(exchange2);
        gate->macroParents.insert((*exchange1->origin->children.begin()));
        exchange1->origin->macroOrigin = nullptr;

        updateSupports(exchange1->origin);

        for (auto& p : parsWithVar) {
            bool found = false;
            if (p.second.find(exchange1) != p.second.end()) {
                p.second.erase(exchange1);
                found = true;
            }
            if (p.second.find(exchange2) != p.second.end()) {
                p.second.erase(exchange2);
                found = true;
            }
            if (found) p.second.insert(*exchange1->origin->children.begin());
        }

    }  // while still more than one parent

    updateSupports(gate);

}  // end restructureMacroParents



unsigned int
HenkinSolver::findSmallestVarOR(SimpleGate* gate, std::vector<Variable>& vars)
{
    unsigned int smallestVar = vars[0];  
    unsigned int fewestParents;

    fewestParents = collectParentsOR(gate, smallestVar);  // should be universal
                                                          // return 0 = there are ex vars
                                                          // that should be pushed first

    unsigned int erase = 0;

    for (size_t v = 1; v < vars.size(); ++v) {
        unsigned int pars = collectParentsOR(gate, vars[v]);

        if (pars == 0) {
        	continue;
        }

        if ((pars < fewestParents) || (fewestParents == 0)) {
            smallestVar   = vars[v];
            fewestParents = pars;
            erase         = v;
        }
    }  // end for


    if (fewestParents == 0) { 
    	return 0;
    }
    if (fewestParents >= gate->macroParents.size()) {
    	return 0;
    }

    vars.erase(vars.begin() + erase);
    return smallestVar;

}  // end findSmallestVarOR



/**
 * \brief restructure the macroGate such that there is only one parent for this var to enable pushing
 */
void
HenkinSolver::restructureMacroParentsUnivOR(SimpleGate* gate, Variable var)
{
    std::vector<SimpleGate*> orderedGates;

    std::set<SimpleEdge*>& pars = parsWithVar.find(var)->second;

    addDepPars(gate, var, pars);

    if (pars.size() == 1) return;

    if (pars.size() == gate->macroParents.size()) return;

    while (pars.size() > 1) {
        findTopoOrder(gate, orderedGates);

        SimpleEdge* exchange1 = NULL;
        SimpleEdge* exchange2 = NULL;

        for (int i = orderedGates.size() - 1; i >= 0; --i) {
            // found a gate where both parents are parents with var
            if ((pars.find(orderedGates[i]->parent1) != pars.end())
                && (pars.find(orderedGates[i]->parent2) != pars.end())) {
              
                exchange1 = orderedGates[i]->parent1;
                exchange2 = orderedGates[i]->parent2;
                break;
            }
            // only found one parent with par with var
            else if (pars.find(orderedGates[i]->parent1) != pars.end()) {
                if (exchange1 == NULL) {
                    exchange1 = orderedGates[i]->parent1;
                } else {
                    exchange2 = orderedGates[i]->parent1;
                    break;
                }
            } else if (pars.find(orderedGates[i]->parent2) != pars.end()) {
                if (exchange1 == NULL) {
                    exchange1 = orderedGates[i]->parent2;
                } else {
                    exchange2 = orderedGates[i]->parent2;
                    break;
                }
            }

        }  // end for current orderedGates

        if ((exchange1 != NULL) && (exchange2 != NULL)) {
            if (exchange1->origin != exchange2->origin) {
                swapParents(exchange1, exchange2);
            }
        } 

        // replace summarized parents by singleparent
        gate->macroParents.erase(exchange1);
        gate->macroParents.erase(exchange2);
        gate->macroParents.insert((*exchange1->origin->children.begin()));
        exchange1->origin->macroOrigin = NULL;

        updateSupports(exchange1->origin);

        for (auto& p : parsWithVar) {
            bool found = false;
            if (p.second.find(exchange1) != p.second.end()) {
                p.second.erase(exchange1);
                found = true;
            }
            if (p.second.find(exchange2) != p.second.end()) {
                p.second.erase(exchange2);
                found = true;
            }
            if (found) p.second.insert(*exchange1->origin->children.begin());
        }

    }  // while still more than one parent

    updateSupports(gate);
}  // end restructureMacroParentsUnivOR






/**
 * var is a universal variable
 * if there are existential variables which we must push first due to dependency, function returns 0
 */
bool
HenkinSolver::isPushableUniv(Variable var)
{
    const std::set<Variable>& deps = _dqbf_prefix->getDependencies(var);  // vars which depend on this universal var

    std::set<SimpleEdge*> intersection;
    for (const auto d : deps) {
        // if we cannot find parsWithVar for this ex variable, then the variable has already been pushed

        // if a parent has d in its support, we must check whether it also has var in its support.
        // if so, then we cannot push var but have to push d first
        if (parsWithVar.find(d) != parsWithVar.end()) {
            std::set<SimpleEdge*>& depsPars = parsWithVar.find(d)->second;
            std::set_intersection(parsWithVar.find(var)->second.begin(), parsWithVar.find(var)->second.end(),
                                  depsPars.begin(), depsPars.end(), std::inserter(intersection, intersection.begin()));

            if (intersection.size() > 0) {
                return false;
            }
        }
    }

    return true;

}  // end isPushableUniv





/**
 * \brief checking which macroParents need to be summarized to push this var.
 * If for a univ var there still are existential variables which must be pushed first
 * return 0.
 * if var is existential, just return number of parsWithVar
 */
unsigned int
HenkinSolver::collectParentsOR(SimpleGate* gate, Variable var)
{
	std::set<SimpleEdge*> pars = parsWithVar.find(var)->second;

    if (_prefix->isUniversal(var)) {
		const std::set<Variable>& deps = _dqbf_prefix->getDependencies(var);  // vars which depend on this universal var

		//std::set<SimpleEdge*> pars = parsWithVar.find(var)->second;  // macroParents with this var in support

		for (const auto d : deps) {
			// if we have d still in parswithvars, then it's not pushed yet and this universal variable has to wait.
			if (parsWithVar.find(d) != parsWithVar.end()) {
				return 0;
			}
			
			// if we cannot find parsWithVar for this variable, then the variable has already been pushed
			for (const auto p : gate->macroParents) {
				for (size_t i = 0; i < p->myGate->supports.size(); ++i) {
					if (!p->myGate->supports[i]) continue;
					if (deps.find(i) != deps.end()) {
						pars.insert(p);
					}
				}
			}
		}
	}

    return pars.size();

}  // end collectParentsOR



/**
 * \brief gate is root of a macrogate
 * find topologic order of all gates in macrogate
 * orderedGates is ordered from root to leafs
 */
void
HenkinSolver::findTopoOrder(SimpleGate* gate, std::vector<SimpleGate*>& orderedGates)
{
    orderedGates.clear();

    std::queue<SimpleGate*> pending;
    std::vector<bool>       seen(maxGateID + 1, false);

    // collect lowest gates in macrogate
    for (const auto p : gate->macroParents) {
        if (seen[p->myGate->id]) continue;
        seen[p->myGate->id] = true;
        for (const auto c : p->myGate->children) {
            if (c->origin->macroOrigin == gate) pending.push(c->origin);
        }
    }

    seen.clear();
    seen.resize(maxGateID + 1, false);

    std::vector<bool> done(maxGateID + 1, false);
    while (!pending.empty()) {
        SimpleGate* currentGate = pending.front();
        pending.pop();

		// only check a gate if already checked both parents
        if (!seen[currentGate->id]) {
            seen[currentGate->id] = true;
            continue;
        }

        if (done[currentGate->id]) continue;

        orderedGates.insert(orderedGates.begin(), currentGate);
        done[currentGate->id] = true;

        for (const auto c : currentGate->children) {
            if (c->origin->macroOrigin == gate) pending.push(c->origin);
        }
    }

}  // end findTopoOrder


/**
 * \brief swap two gates to enable summarize subgraphs below one gate
 * including adjusting of supports of all gates above the modified gates
 */
void
HenkinSolver::swapParents(SimpleEdge* parentStayAtOutGate, SimpleEdge* parentSwapToOutGate)
{
    // parentStayAtOutGate is the parent that is allowed to stay at its gate
    // at the same time its origin is the gate that will be taken out of the macrogate
    SimpleGate* gateKeep = parentSwapToOutGate->origin;
    SimpleGate* gateOut  = parentStayAtOutGate->origin;

    SimpleEdge* parentSwapToKeepGate;
    SimpleEdge* parentStayAtKeepGate;

    // find parentSwapToKeepGate
    if (gateOut->parent1 == parentStayAtOutGate)
        parentSwapToKeepGate = gateOut->parent2;
    else
        parentSwapToKeepGate = gateOut->parent1;

    // find parentStayAtKeepGate
    if (gateKeep->parent1 == parentSwapToOutGate)
        parentStayAtKeepGate = gateKeep->parent2;
    else
        parentStayAtKeepGate = gateKeep->parent1;

    // swap Parents
    if (gateOut->parent1 == parentStayAtOutGate)
        gateOut->parent2 = parentSwapToOutGate;
    else
        gateOut->parent1 = parentSwapToOutGate;

    if (gateKeep->parent1 == parentStayAtKeepGate)
        gateKeep->parent2 = parentSwapToKeepGate;
    else
        gateKeep->parent1 = parentSwapToKeepGate;

    parentSwapToKeepGate->origin = gateKeep;
    parentSwapToOutGate->origin  = gateOut;

    std::vector<SimpleGate*> adjustParents;
    adjustParents.push_back(gateOut);
    adjustParents.push_back(gateKeep);
    adjustSupports(adjustParents);

}  // end swapParents


void
HenkinSolver::updateSupports(SimpleGate* gate)
{
    if (gate->macroOrigin != NULL) {
        std::queue<SimpleGate*> pending;
        seenGates.clear();
        seenGates.resize(maxGateID + 1, false);

        for (const auto p : gate->macroParents) {
            pending.push(p->origin);
        }

        while (!pending.empty()) {
            SimpleGate* currentGate = pending.front();
            pending.pop();

            if (!seenGates[gate->id]) {
                seenGates[gate->id] = true;
                continue;
            }

            currentGate->supports.clear();
            currentGate->supports.resize(maxVarIndex + 1);

            for (size_t i = 1; i <= maxVarIndex; ++i) {
                currentGate->supports[i]
                    = currentGate->parent1->myGate->supports[i] || currentGate->parent2->myGate->supports[i];
            }

            if (currentGate != rootGate) {
                for (const auto c : currentGate->children) {
                    // actually should be only one child
                    pending.push(c->origin);
                }
            }
        }
    } else {
        gate->supports.clear();
        gate->supports.resize(maxVarIndex + 1, false);
        {
            for (size_t i = 1; i <= maxVarIndex; ++i) {
                gate->supports[i] = gate->parent1->myGate->supports[i] || gate->parent2->myGate->supports[i];
            }
        }
    }
}  // end updateSupports



/** 
 * \brief collect all macroparents of gate with var in support
 */
void
HenkinSolver::addDepPars(SimpleGate* gate, Variable var, std::set<SimpleEdge*>& parsInclDeps)
{
    const std::set<Variable>& deps = _dqbf_prefix->getDependencies(var);
    for (const auto p : gate->macroParents) {
        for (const auto d : deps) {
            if (p->myGate->supports[d]) {
                parsInclDeps.insert(p);
                continue;
            }
        }
    }
}  // end addDepPars



void
HenkinSolver::adjustSupports(const std::vector<SimpleGate*>& gates)
{
    std::queue<SimpleGate*> pending;

    std::vector<bool> needAdjustment(maxGateID + 1, false);
    std::vector<bool> seen(maxGateID + 1, false);

    for (const auto g : gates) pending.push(g);

    // marking all gates which need new support
    while (!pending.empty()) {
        if (needAdjustment[pending.front()->id]) {
            pending.pop();
            continue;
        }
        needAdjustment[pending.front()->id] = true;
        for (auto c : pending.front()->children) pending.push(c->origin);
        pending.pop();
    }  // end while

    for (const auto g : gates) pending.push(g);

    // adjusting support for all marked gates
    while (!pending.empty()) {
        SimpleGate* currentGate = pending.front();
        pending.pop();

        if ((!needAdjustment[currentGate->parent1->myGate->id])
            && (!needAdjustment[currentGate->parent2->myGate->id])) {
            if (seen[currentGate->id]) continue;
            currentGate->supports.clear();
            currentGate->supports.resize(maxVarIndex + 1);
            for (size_t i = 1; i <= maxVarIndex; ++i) {
                currentGate->supports[i]
                    = currentGate->parent1->myGate->supports[i] || currentGate->parent2->myGate->supports[i];
            }
            needAdjustment[currentGate->id] = false;
            for (const auto c : currentGate->children) pending.push(c->origin);

            seen[currentGate->id] = true;
        }
    }  //  end while

}  // end adjustSupports


/** 
 * \brief build AIGSolve AIG for this gate
 */
void
HenkinSolver::buildAIG(SimpleGate* gate, std::vector<std::set<SimpleEdge*>>& varAtEdges)
{
    // get parent edgerefs
    aigpp::EdgeRef p1 = aigvec[gate->parent1->myGate->id];
    aigpp::EdgeRef p2 = aigvec[gate->parent2->myGate->id];

    p1 = p1.notIf(gate->parent1->isNeg);
    p2 = p2.notIf(gate->parent2->isNeg);

    // if vars has been pushed to parents, first quantify if possible
    if (gate->parent1->simplePrefix.size() > 0) {
        quantify(gate->parent1, p1, varAtEdges);
    }
    if (gate->parent2->simplePrefix.size() > 0) {
        quantify(gate->parent2, p2, varAtEdges);
    }

    // build edgeref for this gate
    if (!gate->isOR)  // AND gate
    {
        aigvec[gate->id] = aigpp::And(p1, p2);
    } else {
        aigvec[gate->id] = aigpp::Or(p1, p2);
    }
}

/**
 * \brief quantify if possible
 * else do not change newEdgeRef and put all left over variables to prefixes above
 */
void
HenkinSolver::quantify(SimpleEdge* edge, aigpp::EdgeRef& edgeRef, std::vector<std::set<SimpleEdge*>>& varAtEdges)
{
    std::vector<Variable>& prefix = edge->simplePrefix;

    while (true) {
        bool   again = false;
        size_t index = prefix.size();
        // index == prefix.size() -> only ex vars
        // index == prefix.size()-1 -> only univ vars

        quantableVars.clear();

        // first only check the ex vars
        for (int i = prefix.size() - 1; i >= 0; --i) {
            if (_prefix->isUniversal(prefix[i])) {
                index = i;
                break;
            }

            if (isQuantableE(edge->myGate, prefix[i])) {
                quantableVars.push_back(prefix[i]);
                prefix.erase(prefix.begin() + i);
                --index;
                varAtEdges[prefix[i]].erase(edge);
            }
        }  // end checking ex vars

        // quant out quantable vars
        if (quantableVars.size() > 0) {
            quantGate(quantableVars, edgeRef);

            for (const auto q : quantableVars) {
                if (varAtEdges.size() == 0) {
                    _prefix->removeVar(q);
                }
            }
            quantableVars.clear();
        }

        if (index == prefix.size())  // there were only ex vars
        {
            pushVarsToChildren(edge, varAtEdges);
            break;
        }

        // now check all univ vars
        for (int i = index; i >= 0; --i) {
            if (isQuantableA(edge, prefix[i])) {
                quantableVars.push_back(prefix[i]);
                prefix.erase(prefix.begin() + i);
                varAtEdges[i].erase(edge);
            }
        }  // end checking univ vars

        if (quantableVars.size() > 0) {
       
            quantGate(quantableVars, edgeRef);

            for (const auto q : quantableVars) {
                if (varAtEdges.size() == 0) {
                    _prefix->removeVar(q);
                }
            }
            quantableVars.clear();

            again = true;
        }

        if (!again) {
            pushVarsToChildren(edge, varAtEdges);
            break;
        }
    }  // end while
}  // end quantify



/**
 * an existential variable y can only be quantified if the subformula in its scope
 * 1) only contains variable from y's dependency set
 * 2) the dependency sets of all other existential variables y* are subsets of y's dependency set
 */
bool
HenkinSolver::isQuantableE(SimpleGate* gate, Variable var)
{
    const std::set<Variable>& deps = _dqbf_prefix->getDependencies(var);

    for (size_t v = 1; v <= gate->supports.size(); ++v) {
    	if (!gate->supports[v]) continue;

        if (_prefix->isUniversal(v)) {
            if (deps.find(v) == deps.end()) return false;
        }

        if (_prefix->isExistential(v)) {
            const std::set<Variable>& otherDeps = _dqbf_prefix->getDependencies(v);
            for (const auto o : otherDeps) {
                if (deps.find(o) == deps.end()) {
                	return false;
                }
            }
        }
    }

    return true;
}  // end isquantableE

/**
 * a univ var is quantifiable if no ex var at this edge depends on it.
 */
bool
HenkinSolver::isQuantableA(SimpleEdge* edge, Variable var)
{
    const std::set<Variable>& deps = _dqbf_prefix->getDependencies(var);  // vars that depend on this univ var
    if (deps.size() == 0) {
        return true;
    }

    // if one of the variables of the current prefix is dependend on var, var cannot be quantified
    for (int i = edge->simplePrefix.size() - 1; i >= 0; --i) {
        if (_prefix->isUniversal(edge->simplePrefix[i])) break;

        if (deps.find(edge->simplePrefix[i]) != deps.end()) {
            return false;
        }
    }

    return true;
}  // end isQuantableA


/**
 * \brief quantify gate using AIGSolve quantification routine
 */
void
HenkinSolver::quantGate(std::vector<Variable>& quantVars, aigpp::EdgeRef& edgeRef)
{
    _aig_buffer.clear();

    for (const auto var : quantVars) {
        _aig_buffer.push_back(var2aig(var));
    }

    if (_settings.fraig) _manager.makeFRAIG(10);
    if (_settings.rewrite) _manager.rewrite();

    cntQuantifications += _aig_buffer.size();

    _manager.bddQuantifyInplace(edgeRef, _aig_buffer, _prefix->isExistential(quantVars[0]), false, true);

    if (_settings.fraig) _manager.makeFRAIG(10);
    if (_settings.rewrite) _manager.rewrite();

    _aig_buffer.clear();

    if (aigpp::functionallyFalse(edgeRef)) {
        edgeRef = _manager.getConst0();
        return;
    } else if (aigpp::functionallyTrue(edgeRef)) {
        edgeRef = _manager.getConst1();
        return;
    }

}  // end quantGate




/**
 * \brief drag back unquantified vars back towards root
 */
void
HenkinSolver::pushVarsToChildren(SimpleEdge* edge, std::vector<std::set<SimpleEdge*>>& varAtEdges)
{
	for (const auto var : edge->simplePrefix) {
		varAtEdges[var].erase(edge);

		for (auto& c : edge->prefixChildren) {
			bool alreadyThere = false;

			for (auto& cVar : c->simplePrefix) {
				if (var == cVar) {
					alreadyThere = true;
					break;
				}
			}
			if (!alreadyThere) {
				if (_prefix->isExistential(var)) {
					c->simplePrefix.push_back(var);
				}
				else {
					c->simplePrefix.insert(c->simplePrefix.begin(), var);
					c->seenUnivVars.insert(var);
				}

				varAtEdges[var].insert(c);
			}

			c->prefixParents.erase(edge);
		}
	}

	edge->simplePrefix.clear();

}  // end pushVarsToChildren



/*
 * \brief split gate to enable pushing if necessary
 * only gates with more than one child can enter here
 * rootEdge will not enter here
 */
bool
HenkinSolver::pushVarsMacroDQBF2(SimpleEdge* edge, std::vector<SimpleGate*>& newGates)
{
	SimpleGate* gate = edge->myGate;

	std::set<SimpleEdge*> emptyEdges;

	// check for empty edges. They preven variables from being pushed
	for (const auto c : gate->children) {
		if (c->simplePrefix.size() == 0) emptyEdges.insert(c);
	}	

	if (emptyEdges.size() == gate->children.size()) {
		return false;	
	}

	// split away empty edges
	if (emptyEdges.size() > 0) {
		SimpleGate* newGate = splitMacroGate(gate,emptyEdges);
		newGates.push_back(newGate);
		seenChildren.resize(maxGateID+1,false);
	}
	
	std::vector<Variable> commonVars;

	if (gate->children.size() == 1) {
		commonVars = (*gate->children.begin())->simplePrefix;
		return pushVarsMacroDQBF(*gate->children.begin(), commonVars);
	}

	bool samePrefixes = sameEdges(gate);

	if (samePrefixes) {
		commonVars = (*gate->children.begin())->simplePrefix;
		return pushVarsMacroDQBF(*gate->children.begin(), commonVars);
	}

	if (gate->isOR) {
		return pushVarsMacroDQBFORSplit(gate, newGates);
	}
	else {
		return pushVarsMacroDQBFANDSplit(gate, newGates);
	}
} //  end pushVarsMacroDQBF2


bool
HenkinSolver::pushVarsDQBF2(SimpleEdge* edge, std::vector<SimpleGate*>& newGates)
{
	SimpleGate* gate = edge->myGate;

	std::set<SimpleEdge*> emptyEdges;

	// check for empty edges. They preven variables from other edges being pushed
	for (const auto c : gate->children) {
		if (c->simplePrefix.size() == 0) emptyEdges.insert(c);
	}	

	if (emptyEdges.size() == gate->children.size()) {
		return false;	
	}

	// split away empty edges
	if (emptyEdges.size() > 0) {
		SimpleGate* newGate = splitGate(gate,emptyEdges);
		newGates.push_back(newGate);
		seenChildren.resize(maxGateID+1,false);
	}

	std::vector<Variable> commonVars;

	if (gate->children.size() == 1) {
		commonVars = (*gate->children.begin())->simplePrefix;
		return pushVarsDQBF(*gate->children.begin(), commonVars);
	}

	bool samePrefixes = sameEdges(gate);	
	
	if (samePrefixes) {
		commonVars = (*gate->children.begin())->simplePrefix;
		return pushVarsDQBF(*gate->children.begin(), commonVars);
	}

	if (gate->isOR) {
		return pushVarsDQBFORSplit(gate, newGates);
	}
	else {
		return pushVarsDQBFANDSplit(gate, newGates);
	}
} //  end pushVarsDQBF2





bool
HenkinSolver::pushVarsMacroDQBFORSplit(SimpleGate* gate, std::vector<SimpleGate*>& newGates)
{
	bool pushed = false;

	// when a gate gets into pending, it needs splitting
	std::map<Variable, std::set<SimpleEdge*>> varsAtChildren;
	//std::vector<Variable> occurringVars;
	//std::vector<Variable> commonVars;
	//std::set<Variable> stuckedVars;
	std::vector<Variable> varsVec; // use this in all kind of situations
	//std::set<Variable> pushedVars; 
	std::set<SimpleEdge*> emptyEdges;

	std::queue<SimpleGate*> pending;
	pending.push(gate);

	while (!pending.empty()) { // repeat if gate has been split
		SimpleGate* currentGate = pending.front();
		pending.pop();
		
		getVarsAtChildren(currentGate, varsAtChildren/*, occurringVars*/);
		for (const auto v : varsAtChildren) varsVec.push_back(v.first); 
		getParsWithVar(currentGate, varsVec);
		varsVec.clear();

		// try to push existential variables
		for (const auto v : varsAtChildren) {
			if (!_prefix->isExistential(v.first)) continue;
			if (v.second.size() < currentGate->children.size()) continue;
			if (!isPushableEx(parsWithVar.find(v.first)->second, v.first, currentGate)) continue;

			deleteVarFromChildren(currentGate, v.first);
			distributeVarsMacro(currentGate, v.first);
			parsWithVar.erase(v.first);
			varsVec.push_back(v.first);
		}

		if (varsVec.size() > 0) {
			for (const auto v : varsVec) varsAtChildren.erase(v);
			varsVec.clear();
			pushed = true;
		}


		
		// collect pushable universal variables
		for (const auto v : varsAtChildren) {
			if (!_prefix->isUniversal(v.first)) continue;
			if (v.second.size() < currentGate->children.size()) continue;

			const auto dep = _dqbf_prefix->getDependencies(v.first);
			bool stucked = false;
			for (const auto d : dep) {
				if (varsAtChildren.find(d) != varsAtChildren.end()) {
					stucked = true;
					break;
				}
			}

			if (!stucked) varsVec.push_back(v.first);
		}

		// try to push universal variables
		while (!varsVec.empty()) {
			Variable nextVar = findSmallestVarOR(currentGate, varsVec);

			if (nextVar == 0) break;
			if (parsWithVar.find(nextVar)->second.size() == currentGate->macroParents.size()) break;

			// restructure gate for pushing
			if (_prefix->isUniversal(nextVar)) restructureMacroParentsUnivOR(currentGate, nextVar);
			else restructureMacroParents(currentGate, nextVar);
			
			// push variable
			if (_prefix->isUniversal(nextVar)) {
				(*(parsWithVar.find(nextVar)->second).begin())
					->simplePrefix.insert((*(parsWithVar.find(nextVar)->second).begin())->simplePrefix.begin(),
									  nextVar);          }
			else {
				(*(parsWithVar.find(nextVar)->second).begin())->simplePrefix.push_back(nextVar);
			}

			// do rest
			deleteVarFromChildren(currentGate, nextVar);
			varsAtChildren.erase(nextVar);
			parsWithVar.erase(nextVar);
			pushed = true;
		}

		varsAtChildren.clear();
		varsVec.clear();
		//occurringVars.clear();
		//stuckedVars.clear();
		//pushableVars.clear();
		//commonVars.clear();

		
		// checking for empty edges 
		for (const auto c : currentGate->children) {
			if (c->simplePrefix.size() == 0) emptyEdges.insert(c);
		}

		if (emptyEdges.size() == currentGate->children.size()) {
			return pushed;
		}
		else if (emptyEdges.size() > 0) {
			SimpleGate* newGate = splitMacroGate(currentGate,emptyEdges);
			newGates.push_back(newGate);
			seenChildren.resize(maxGateID+1,false);
			pending.push(currentGate);
		}
		else {
			getVarsAtChildren(currentGate, varsAtChildren/*, occurringVars*/);
			
			//for (const auto v : varsAtChildren) occurringVars.push_back(v.first);
			//getParsWithVar(currentGate, occurringVars);

			Variable varToSplit = 0;
	
			// TODO: find heuristics to choose var for splitting
			// now: first try existential ones, then universal ones, simply take the first one we get
			for (const auto v : varsAtChildren) {
				if (!_prefix->isExistential(v.first)) continue;
				if (v.second.size() == currentGate->children.size()) continue;	
				
				varToSplit = v.first;
			}
			
			if (varToSplit == 0) {
				for (const auto v : varsAtChildren) {
					if (!_prefix->isUniversal(v.first)) continue;
					if (v.second.size() == currentGate->children.size()) continue;	
					
					varToSplit = v.first;
				}	
			}

			if (varToSplit == 0) continue;

			SimpleGate* newGate = splitMacroGate(currentGate, varsAtChildren.find(varToSplit)->second);
			newGates.push_back(newGate);
			seenChildren.resize(maxGateID+1,false);
			pending.push(newGate);
			pending.push(currentGate);

			varsAtChildren.clear();
			emptyEdges.clear();

		} 
	} // end while for all gates

	return pushed;

} // end pushVarsMacroDQBFORSplit




bool
HenkinSolver::pushVarsMacroDQBFANDSplit(SimpleGate* gate, std::vector<SimpleGate*>& newGates)
{
	// rootEdge will not enter here
	// all edges are carrying variables
	// at least two of them are different

	bool pushed = false;
	
	std::map<Variable, std::set<SimpleEdge*>> varsAtChildren;
	std::vector<Variable> occurringVars;
	std::vector<Variable> pushableUnivVars;
	std::vector<Variable> pushableExVars;
	std::set<SimpleEdge*> emptyEdges;
	std::vector<Variable> commonVars;

	std::queue<SimpleGate*> pending;
	pending.push(gate);

	while (!pending.empty()) {
		SimpleGate* currentGate = pending.front();
		pending.pop();
		
		getVarsAtChildren(currentGate, varsAtChildren);
		for (const auto v : varsAtChildren) occurringVars.push_back(v.first);
		getParsWithVar(currentGate, occurringVars);

		// collect variables which can be pushed
		for (const auto v : varsAtChildren) {
			if (v.second.size() == currentGate->children.size()) {
				if (_prefix->isExistential(v.first)) {
					if (parsWithVar.find(v.first)->second.size() < currentGate->macroParents.size())
						pushableExVars.push_back(v.first);
				}
				else
					pushableUnivVars.push_back(v.first);
			}
		}

		
		// try to push universal variables which is not blocked by any existential variable
		for (const auto v : pushableUnivVars) {
			bool stuck = false;
			const auto dep = _dqbf_prefix->getDependencies(v);
			for (const auto d : dep) {
				if (varsAtChildren.find(d) != varsAtChildren.end()) {
					stuck = true;
					break;
				}
			}

			if (stuck) continue;
			
			distributeVarsMacro(currentGate, v);
			deleteVarFromChildren(currentGate, v);
			varsAtChildren.erase(v);
			parsWithVar.erase(v);
			pushed = true;
		}

		// try to push ex vars and univ vars alternatingly
		while (true) {
			bool pushedVar = false;

			// find pushable existential variables
			while (true) {
				Variable nextVar = 0;

				for (const auto v : pushableExVars) {
					if (nextVar == 0) nextVar = v;
					if (parsWithVar.find(nextVar)->second.size() > parsWithVar.find(v)->second.size())
						nextVar = v;
				}

				if (nextVar == 0) break;
				if (parsWithVar.find(nextVar)->second.size() == currentGate->macroParents.size()) break;

				restructureMacroParents(currentGate, nextVar);
				(*parsWithVar.find(nextVar)->second.begin())->simplePrefix.push_back(nextVar);
				deleteVarFromChildren(currentGate, nextVar);
				varsAtChildren.erase(nextVar);
				parsWithVar.erase(nextVar);
				pushedVar = true;
				pushed = true;

				for (int i = pushableExVars.size() -1; i >= 0; --i) {
					if (pushableExVars[i] == nextVar) {
						pushableExVars.erase(pushableExVars.begin()+i);
					}
					else if (parsWithVar.find(pushableExVars[i])->second.size() == currentGate->macroParents.size()) {
						pushableExVars.erase(pushableExVars.begin()+i);
					}
				}
			} // while for pushing ex var

			if (!pushedVar) break;

			pushedVar = false;

			// try to push universal variables which is not blocked by any existential variable
			for (int i = pushableUnivVars.size() -1; i >= 0; --i) {
				bool stuck = false;
				const auto dep = _dqbf_prefix->getDependencies(pushableUnivVars[i]);
				for (const auto d : dep) {
					if (varsAtChildren.find(d) != varsAtChildren.end()) {
						stuck = true;
						break;
					}
				}

				if (stuck) continue;
				
				distributeVarsMacro(currentGate, pushableUnivVars[i]);
				deleteVarFromChildren(currentGate, pushableUnivVars[i]);
				varsAtChildren.erase(pushableUnivVars[i]);
				parsWithVar.erase(pushableUnivVars[i]);
				pushableUnivVars.erase(pushableUnivVars.begin()+i);
				pushedVar = true;
				pushed = true;
				
			}

			if (!pushedVar) break;
		} // end while pushing all pushable variables


		// check for empty edges
		for (const auto c : currentGate->children) {
			if (c->simplePrefix.size() == 0) emptyEdges.insert(c);
		}

		if (emptyEdges.size() == currentGate->children.size()) {
			emptyEdges.clear();
			occurringVars.clear();
			varsAtChildren.clear();
			pushableUnivVars.clear();
			pushableExVars.clear();
			continue;	
		}
		// split away empty edges
		else if (emptyEdges.size() > 0) {
			SimpleGate* newGate = splitMacroGate(currentGate,emptyEdges);
			newGates.push_back(newGate);
			seenChildren.resize(maxGateID+1,false);
		
			if (currentGate->children.size() == 1) {
				commonVars = (*currentGate->children.begin())->simplePrefix;
				
				if (pushVarsMacroAnd(currentGate, commonVars)) {
					pushed = true;
				}
				commonVars.clear();
			}
			else {
				pending.push(currentGate);
			}

			emptyEdges.clear();
			occurringVars.clear();
			varsAtChildren.clear();
			pushableUnivVars.clear();
			pushableExVars.clear();
		}
		else { // find var to split and pass both new gates to pending
			Variable nextVar = 0;
			size_t nuChildren = 0;

			// first try to find ex var for splitting
			for (const auto v : varsAtChildren) {
				if (!_prefix->isExistential(v.first)) continue;
				if (v.second.size() == currentGate->children.size()) continue;
				
				if (nextVar == 0) {
					nextVar = v.first;
					nuChildren = v.second.size();
				}
				else if (nuChildren > v.second.size()) {
					nextVar = v.first;
					nuChildren = v.second.size();
				}
			}

			if (nextVar != 0) {
				SimpleGate* newGate = splitMacroGate(currentGate, varsAtChildren.find(nextVar)->second);
				newGates.push_back(newGate);
				seenChildren.resize(maxGateID+1,false);
				pending.push(currentGate);
				pending.push(newGate);

				emptyEdges.clear();
				occurringVars.clear();
				varsAtChildren.clear();
				pushableUnivVars.clear();
				pushableExVars.clear();
				continue;
			}

			// if no ex var for splitting, try to find univ var for splitting
			for (const auto v : varsAtChildren) {
				if (!_prefix->isUniversal(v.first)) continue;
				if (v.second.size() == currentGate->children.size()) continue;
				
				if (nextVar == 0) {
					nextVar = v.first;
					nuChildren = v.second.size();
				}
				else if (nuChildren > v.second.size()) {
					nextVar = v.first;
					nuChildren = v.second.size();
				}
			}
			
			if (nextVar != 0) {
				SimpleGate* newGate = splitMacroGate(currentGate, varsAtChildren.find(nextVar)->second);
				newGates.push_back(newGate);
				seenChildren.resize(maxGateID+1,false);
				pending.push(currentGate);
				pending.push(newGate);

				emptyEdges.clear();
				occurringVars.clear();
				varsAtChildren.clear();
				pushableUnivVars.clear();
				pushableExVars.clear();
			}
		}
	} // end while (!pending.empty())


	return pushed;
} // end pushVarsMacroDQBFANDSplit




bool
HenkinSolver::pushVarsDQBFORSplit(SimpleGate* gate, std::vector<SimpleGate*>& newGates)
{
	bool pushed = false;

	// when a gate gets into pending, it needs splitting
	std::map<Variable, std::set<SimpleEdge*>> varsAtChildren;
	//std::vector<Variable> occurringVars;
	//std::vector<Variable> commonVars;
	//std::set<Variable> stuckedVars;
	std::vector<Variable> pushedVars; 
	std::set<Variable> unpushableVars; // these variables cannot be pushed, no matter how we split the gate

	std::queue<SimpleGate*> pending;
	pending.push(gate);

	while (!pending.empty()) { // repeat if gate has been split
		SimpleGate* currentGate = pending.front();
		pending.pop();
		
		getVarsAtChildren(currentGate, varsAtChildren/*, occurringVars*/);
		//getParsWithVar(currentGate, occurringVars);

		// try to push all existential variables which occur in all children edges
		for (const auto v : varsAtChildren) {
			if (unpushableVars.find(v.first) != unpushableVars.end()) continue;
			if (!_prefix->isExistential(v.first)) continue;
			if (v.second.size() < currentGate->children.size()) continue;

			if (currentGate->parent1->myGate->supports[v.first]) {
				if (currentGate->parent2->myGate->supports[v.first]) {
					std::set<SimpleEdge*> parents;
					parents.insert(currentGate->parent1);
					parents.insert(currentGate->parent2);
					if (isPushableEx(parents, v.first, currentGate)) {
						currentGate->parent1->simplePrefix.push_back(v.first);
						currentGate->parent2->simplePrefix.push_back(v.first);
						deleteVarFromChildren(currentGate, v.first);
						//varsAtChildren.erase(v.first);
						pushedVars.push_back(v.first);
						//pushed = true;
					}
					else {
						unpushableVars.insert(v.first);
					}
				}
				else {
					currentGate->parent1->simplePrefix.push_back(v.first);
					deleteVarFromChildren(currentGate, v.first);
					//varsAtChildren.erase(v.first);
					//pushed = true;
					pushedVars.push_back(v.first);
				}
			}
			else {
				currentGate->parent2->simplePrefix.push_back(v.first);
				deleteVarFromChildren(currentGate, v.first);
				//varsAtChildren.erase(v.first);
				//pushed = true;
				pushedVars.push_back(v.first);
			}
		}

		if (pushedVars.size() > 0) {
			pushed = true;
			for (const auto p : pushedVars) varsAtChildren.erase(p);
			pushedVars.clear();
		}
		
		// existential variables left in varAtChildren are only not pushable in this iteration
		// but not ruled out for pushing per se

		// collect pushable universal variables
		for (const auto v : varsAtChildren) {
			if (!_prefix->isUniversal(v.first)) continue;
			if (unpushableVars.find(v.first) != unpushableVars.end()) continue;

			if (currentGate->parent1->myGate->supports[v.first]) {
				if (!currentGate->parent2->myGate->supports[v.first]) {
					if (v.second.size() < currentGate->children.size()) continue;

					const auto dep = _dqbf_prefix->getDependencies(v.first);
					bool stucked = false;
					for (const auto d : dep) {
						if (varsAtChildren.find(d) != varsAtChildren.end()) {
							stucked = true;
							break;
						}
					}

					if (stucked) continue;

					currentGate->parent1->simplePrefix.insert(currentGate->parent1->simplePrefix.begin(),v.first);
					deleteVarFromChildren(currentGate, v.first);
					pushedVars.push_back(v.first);
				}
				else {
					unpushableVars.insert(v.first);
				}
			}
			else
			{
				if (v.second.size() < currentGate->children.size()) continue;
			
				const auto dep = _dqbf_prefix->getDependencies(v.first);
				bool stucked = false;
				for (const auto d : dep) {
					if (varsAtChildren.find(d) != varsAtChildren.end()) {
						stucked = true;
						break;
					}
				}

				if (stucked) continue;				

				currentGate->parent2->simplePrefix.insert(currentGate->parent2->simplePrefix.begin(),v.first);
				deleteVarFromChildren(currentGate, v.first);
				//varsAtChildren.erase(v.first);
				//pushed = true;
				pushedVars.push_back(v.first);
			}
		}

		if (pushedVars.size() > 0) {
			for (const auto p : pushedVars) varsAtChildren.erase(p);
			pushedVars.clear();
			pushed = true;
		}
		
		varsAtChildren.clear();

		// checking for empty edges 
		std::set<SimpleEdge*> emptyEdges;
		for (const auto c : currentGate->children) {
			if (c->simplePrefix.size() == 0) emptyEdges.insert(c);
		}

		if (emptyEdges.size() == currentGate->children.size()) {
			return pushed;
		}
		else if (emptyEdges.size() > 0) {
			SimpleGate* newGate = splitGate(currentGate,emptyEdges);
			newGates.push_back(newGate);
			seenChildren.resize(maxGateID+1,false);
			pending.push(currentGate);
		}
		else { // no emptyedges
			getVarsAtChildren(currentGate, varsAtChildren/*, occurringVars*/);

			Variable varToSplit = 0;
	
			// TODO: find heuristics to choose var for splitting
			// now: first try existential ones, then universal ones, simply take the first one we get
			for (const auto v : varsAtChildren) {
				if (!_prefix->isExistential(v.first)) continue;
				if (v.second.size() == currentGate->children.size()) continue;	
				if (unpushableVars.find(v.first) != unpushableVars.end()) continue;
				
				varToSplit = v.first;
			}

			if (varToSplit == 0) {
				for (const auto v : varsAtChildren) {
					if (_prefix->isUniversal(v.first)) continue;
					if (v.second.size() == currentGate->children.size()) continue;	
					if (unpushableVars.find(v.first) != unpushableVars.end()) continue;
					
					varToSplit = v.first;
				}	
			}

			if (varToSplit == 0) continue;

			SimpleGate* newGate = splitGate(currentGate, varsAtChildren.find(varToSplit)->second);
			newGates.push_back(newGate);
			seenChildren.resize(maxGateID+1,false);
			pending.push(newGate);
			pending.push(currentGate);
			varsAtChildren.clear();
		}
	} // end while for all gates

	return pushed;


} // end pushVarsDQBFORSplit



bool
HenkinSolver::pushVarsDQBFANDSplit(SimpleGate* gate, std::vector<SimpleGate*>& newGates)
{
// rootEdge will not enter here
// all edges are carrying variables
// at least two of them are different

	bool pushed = false;

	std::map<Variable, std::set<SimpleEdge*>> varsAtChildren;
	std::set<Variable> unpushableVars; // variables which cannot be pushed even if we split
	std::vector<Variable> commonVars;
	getVarsAtChildren(gate, varsAtChildren);


	std::vector<Variable> varsToDelete;

	// if an ex var occur in all incoming edges but also in both parentsupports
	// it will not become pushable due to splitting or
	// free the way for some univ vars due to splitting
	// we can compute all of them in the first iteration

	// an ex var which occurs in both parents but not in all children is not considered unpushable because for 
	// x     y(x)
	//  \   /
	//  gate
	// -> splitting gate enables pushing of x

	// finding unpushable ex vars
	// if an existential variable can be pushed, push right away
	for (const auto v : varsAtChildren) {
		if (!_prefix->isExistential(v.first)) continue;

		if (v.second.size() < gate->children.size()) continue;

		if (gate->parent1->myGate->supports[v.first]) {
			if (gate->parent2->myGate->supports[v.first]) {
				unpushableVars.insert(v.first);
			}
			else {
				gate->parent2->simplePrefix.push_back(v.first);
				deleteVarFromChildren(gate, v.first);
				varsToDelete.push_back(v.first);
				pushed = true;
			} 	
		}
		else {
			gate->parent1->simplePrefix.push_back(v.first);
			deleteVarFromChildren(gate, v.first);
			varsToDelete.push_back(v.first);
			pushed = true;
		}
	}

	for (const auto v : varsToDelete) varsAtChildren.erase(v);

	varsToDelete.clear();

	// finding unpushable univ vars
	for (const auto v : varsAtChildren) {
		if (!_prefix->isUniversal(v.first)) continue;

		// if a univ var is blocked by an ex var which cannot be pushed, it is also stuck
		const auto dep = _dqbf_prefix->getDependencies(v.first);
		bool stuck = false;
		for (const auto d : dep) {
			if (unpushableVars.find(d) != unpushableVars.end()) {
				unpushableVars.insert(v.first);
				stuck = true;
				break;
			}
			if (varsAtChildren.find(d) != varsAtChildren.end()) {
				// cannot be pushed right now, but maybe later
				stuck = true;
				break;
			}
		}

		if (stuck) continue;
		
		if (v.second.size() < gate->children.size()) continue;
		
		// if not stuck, can be pushed right away
		if (gate->parent1->myGate->supports[v.first]) {
			gate->parent1->simplePrefix.insert(gate->parent1->simplePrefix.begin(), v.first);
		}
		if (gate->parent2->myGate->supports[v.first]) {
			gate->parent2->simplePrefix.insert(gate->parent2->simplePrefix.begin(), v.first);
		}

		deleteVarFromChildren(gate, v.first);
		varsToDelete.push_back(v.first);
		pushed = true;

	}

	for (const auto v : varsToDelete) varsAtChildren.erase(v);
	varsAtChildren.clear();

	std::set<SimpleEdge*> emptyEdges;
	std::set<Variable> currentUnpushableVars;

	std::queue<SimpleGate*> pending;
	pending.push(gate);

	while (!pending.empty()) {

		SimpleGate* currentGate = pending.front();
		pending.pop();
			
		for (const auto c : currentGate->children) {
			if (c->simplePrefix.size() == 0) emptyEdges.insert(c);
		}

		if (emptyEdges.size() == currentGate->children.size()) {
			emptyEdges.clear();
			continue;
		}

		if (emptyEdges.size() > 0 ) {
			SimpleGate* newGate = splitGate(currentGate, emptyEdges);
			newGates.push_back(newGate);
			emptyEdges.clear();
		}

		if (currentGate->children.size() == 1) {
			commonVars = (*currentGate->children.begin())->simplePrefix;
			if (pushVarsDQBF(*currentGate->children.begin(), commonVars)) {
				pushed = true;
			}
			continue;
		}

		if (sameEdges(currentGate)) {
			commonVars = (*currentGate->children.begin())->simplePrefix;
			if (pushVarsDQBF(*currentGate->children.begin(), commonVars)) {
				pushed = true;
			}
			continue;
		}

		// there is more than one edge and at least two of them have diffent variables


		getVarsAtChildren(currentGate, varsAtChildren);
		currentUnpushableVars = unpushableVars;

		bool pushedInThisLoop = false;

		// finding unpushable ex vars or push ex vars which can be pushed
		for (const auto v : varsAtChildren) {
			if (!_prefix->isExistential(v.first)) continue;
			if (v.second.size() < currentGate->children.size()) continue;
			if (currentUnpushableVars.find(v.first) != currentUnpushableVars.end()) continue;

			if (currentGate->parent1->myGate->supports[v.first]) {
				if (currentGate->parent2->myGate->supports[v.first]) {
					currentUnpushableVars.insert(v.first);
				}
				else {
					currentGate->parent2->simplePrefix.push_back(v.first);
					deleteVarFromChildren(currentGate, v.first);
					varsToDelete.push_back(v.first);
					pushed = true;
					pushedInThisLoop = true;
				} 	
			}
			else {
				currentGate->parent1->simplePrefix.push_back(v.first);
				deleteVarFromChildren(currentGate, v.first);
				varsToDelete.push_back(v.first);
				pushed = true;
				pushedInThisLoop = true;
			}
		}

		for (const auto v : varsToDelete) varsAtChildren.erase(v);
		varsToDelete.clear();

		// finding unpushable univ vars or push ex vars which can be pushed

		for (const auto v : varsAtChildren) {
			if (!_prefix->isUniversal(v.first)) continue;
			if (v.second.size() < currentGate->children.size()) continue;
			if (currentUnpushableVars.find(v.first) != currentUnpushableVars.end()) continue;

			// if a univ var is blocked by an ex var which cannot be pushed, it is also stuck
			const auto dep = _dqbf_prefix->getDependencies(v.first);
			bool stuck = false;
			for (const auto d : dep) {
				if (currentUnpushableVars.find(d) != currentUnpushableVars.end()) {
					currentUnpushableVars.insert(v.first);
					stuck = true;
					break;
				}
				if (varsAtChildren.find(d) != varsAtChildren.end()) {
					stuck = true;
				}
			}

			if (stuck) continue;

			if (currentGate->parent1->myGate->supports[v.first]) {
				currentGate->parent1->simplePrefix.insert(currentGate->parent1->simplePrefix.begin(), v.first);
			}
			if (currentGate->parent2->myGate->supports[v.first]) {
				currentGate->parent2->simplePrefix.insert(currentGate->parent2->simplePrefix.begin(), v.first);
			}

			deleteVarFromChildren(currentGate, v.first);
			varsToDelete.push_back(v.first);
			pushed = true;
			pushedInThisLoop = true;

		}

		for (const auto v : varsToDelete) varsAtChildren.erase(v);
		varsToDelete.clear();

		if (pushedInThisLoop) {
			pending.push(currentGate);
			varsAtChildren.clear();
			currentUnpushableVars.clear();
			continue;
		}

		Variable varToSplit = 0;
		
		// try to find ex var which occur in the fewest children
		for (const auto v : varsAtChildren) {
			if (!_prefix->isExistential(v.first)) continue;
			if (currentUnpushableVars.find(v.first) != currentUnpushableVars.end()) continue;

			if (varToSplit == 0) {
				varToSplit = v.first;
			}
			else if (varsAtChildren.find(varToSplit)->second.size() < v.second.size()) {
				varToSplit = v.first;
			}
		}

		// no ex var found, take first univ var to push
		if (varToSplit == 0) {
			for (const auto v : varsAtChildren) {
				if (!_prefix->isUniversal(v.first)) continue;
				if (currentUnpushableVars.find(v.first) != currentUnpushableVars.end()) continue;

				varToSplit = v.first;
				break;				
			}
		}
		
		if (varToSplit == 0) return pushed;


		SimpleGate* newGate = splitGate(currentGate, varsAtChildren.find(varToSplit)->second);
		newGates.push_back(newGate);
		pending.push(currentGate);
		pending.push(newGate);
		
		varsAtChildren.clear();
		currentUnpushableVars.clear();
	} // end for all gates

	return pushed;

} // end pushVarsDQBFANDSplit




/*
 * collect all variables which occur in any of the incoming edges
 * collect for each variable the edges it occurs at
 * no edge coming in here is empty
 */
void
HenkinSolver::getVarsAtChildren(SimpleGate* gate, std::map<Variable,std::set<SimpleEdge*>>& varsAtChildren)
{
	for (const auto c : gate->children) {
		for (const auto v : c->simplePrefix) {
			if (varsAtChildren.find(v) == varsAtChildren.end()) {
				std::set<SimpleEdge*> tmp;
				tmp.insert(c);
				varsAtChildren[v] = tmp;
			}
			else
				varsAtChildren.find(v)->second.insert(c);
		}
	}
} // end getVarAtChildrenAndOccurringVariables




/*
 * \brief splits a gate to separate the children
 * only apply if gate is not a variable! 
 * newGate will carry childrenForNewGate
 */
HenkinSolver::SimpleGate*
HenkinSolver::splitGate(SimpleGate* origGate, std::set<SimpleEdge*>& childrenForNewGate)
{
	SimpleGate* newGate = new SimpleGate();

	++maxGateID;
	newGate->id = maxGateID;
	newGate->isOR = origGate->isOR;
	newGate->supports = origGate->supports;

	SimpleEdge* newParent1 = new SimpleEdge();
	newParent1->isNeg = origGate->parent1->isNeg;
	newParent1->myGate = origGate->parent1->myGate;
	newParent1->origin = newGate;
	newParent1->simplePrefix = origGate->parent1->simplePrefix;
	newGate->parent1 = newParent1;
	newParent1->myGate->children.insert(newParent1);

	SimpleEdge* newParent2 = new SimpleEdge();
	newParent2->isNeg = origGate->parent2->isNeg;
	newParent2->myGate = origGate->parent2->myGate;
	newParent2->origin = newGate;
	newParent2->simplePrefix = origGate->parent2->simplePrefix;
	newGate->parent2 = newParent2;
	newParent2->myGate->children.insert(newParent2);
	
	for (auto c : childrenForNewGate) {
		origGate->children.erase(c);
		c->myGate = newGate;
	}
	newGate->children = childrenForNewGate;	

	return newGate;

} // end splitGate
	
	
HenkinSolver::SimpleGate* 
HenkinSolver::splitMacroGate(SimpleGate* origGate, std::set<SimpleEdge*>& childrenForNewGate)
{
	SimpleGate* newGate = splitGate(origGate, childrenForNewGate);
	newGate->macroOrigin = newGate;

	std::queue<SimpleGate*> pending;

	if (origGate->parent1->myGate->macroOrigin == origGate) {
		pending.push(origGate->parent1->myGate);
	}
	else {
		for (const auto c : origGate->parent1->myGate->children) {
			if (c->origin == newGate) {
				newGate->macroParents.insert(c);
				break;
			}
		}
	}

	if (origGate->parent2->myGate->macroOrigin == origGate) {
		pending.push(origGate->parent2->myGate);
	}
	else {
		for (const auto c : origGate->parent2->myGate->children) {
			if (c->origin == newGate) {
				newGate->macroParents.insert(c);
				break;
			}
		}
	}

	childrenForNewGate.clear();

	while (!pending.empty()) {
		SimpleGate* currentGate = pending.front();
		pending.pop();
		if (currentGate->children.size() != 2) {
			std::cout << "not two children! Something wrong" << std::endl;
			exit(0);
		}


		std::set<SimpleEdge*>::iterator iter = currentGate->children.begin();
		SimpleEdge* child1 = *iter;
		++iter;
		SimpleEdge* child2 = *iter;

		if (child1->origin->id > child2->origin->id) childrenForNewGate.insert(child1);
		else childrenForNewGate.insert(child2);

		SimpleGate* currentNewGate = splitGate(currentGate, childrenForNewGate);
		currentNewGate->macroOrigin = newGate;

		if (currentGate->parent1->myGate->macroOrigin == origGate) {
			pending.push(currentGate->parent1->myGate);
		}
		else {
			for (const auto c : currentGate->parent1->myGate->children) {
				if (c->origin == currentNewGate) {
					newGate->macroParents.insert(c);
					break;
				}
			}
		}
		if (currentGate->parent2->myGate->macroOrigin == origGate) {
			pending.push(currentGate->parent2->myGate);
		}
		else {
			for (const auto c : currentGate->parent2->myGate->children) {
				if (c->origin == currentNewGate) {
					newGate->macroParents.insert(c);
					break;
				}
			}
		}

		childrenForNewGate.clear();
	}

	return newGate;

} // end splitMacroGate


/*
 * \brief checks whether all incomingEdges carry the same variables
 */
bool 
HenkinSolver::sameEdges(SimpleGate* gate)
{
	// check whether all variables occur in all edges
	std::vector<unsigned int> cntVariables(maxVarIndex+1, 0);
	for (const auto c : gate->children) {
		for (const auto v : c->simplePrefix) ++cntVariables[v];
	}
	bool sameEdges = true;
	for (size_t i = 1; i < cntVariables.size(); ++i) {
		if (cntVariables[i] == 0) continue;
		if (cntVariables[i] < gate->children.size()) {
			sameEdges = false;
			break;
		}
	}

	return sameEdges;
}


} // end namespace hqs
