#include <fstream>
#include <iostream>
#include <chrono>

#include <boost/iostreams/chain.hpp>
#include <boost/iostreams/filter/gzip.hpp>
#include <boost/iostreams/filtering_stream.hpp>
#include <boost/program_options.hpp>

#include <easylogging++.hpp>
#include "../hqspre/formula.hpp"
#include "hqs.hpp"

#define ELPP_STL_LOGGING

INITIALIZE_EASYLOGGINGPP

static constexpr int result_SAT   = 10;
static constexpr int result_UNSAT = 20;

std::vector<hqspre::Variable>
createQBFPrefix(const hqspre::Formula& formula)
{
    std::vector<hqspre::Variable> order;
    const hqspre::DQBFPrefix* prefix = formula.getDQBFPrefix();
    val_assert(prefix != nullptr);

    order.reserve(prefix->maxVarIndex());

    // Get all existential variables
    std::vector<hqspre::Variable> exist_vars;
    for (hqspre::Variable var = prefix->minVarIndex(); var <= prefix->maxVarIndex(); ++var) {
        if (prefix->isExistential(var)) {
            exist_vars.push_back(var);
        }
    }

    // Sort existential variables according to their number of dependencies.
    // This brings them into the right order for the QBF prefix.
    std::sort(exist_vars.begin(), exist_vars.end(), [prefix](hqspre::Variable a, hqspre::Variable b) {
        const auto deps_a = prefix->numDependencies(a);
        const auto deps_b = prefix->numDependencies(b);
        return (deps_a < deps_b) || (deps_a == deps_b && a < b);
    });

    // Insert the universal variables into the order.
    std::vector<bool> finished_universal(prefix->maxVarIndex() + 1, false);
    for (const auto exist : exist_vars) {
        const auto& deps = prefix->getDependencies(exist);
        for (const hqspre::Variable dep : deps) {
            if (!finished_universal[dep]) {
                finished_universal[dep] = true;
                order.push_back(dep);
            }
        }
        order.push_back(exist);
    }

    // Append the remaining universal variables which do not appear in
    // any dependency sets.
    for (hqspre::Variable univ = prefix->minVarIndex(); univ <= prefix->maxVarIndex(); ++univ) {
        if (prefix->isUniversal(univ) && !finished_universal[univ]) {
            finished_universal[univ] = true;
            order.push_back(univ);
        }
    }

    return order;
}


int
main(int argc, char** argv)
{
	// begin time measure
	std::chrono::high_resolution_clock::time_point beginTime = std::chrono::high_resolution_clock::now();
	std::chrono::high_resolution_clock::time_point solveTime;
	std::chrono::duration<double> duration;

    // Configure logging
    el::Configurations defaultConf;
    defaultConf.setToDefault();
    defaultConf.setGlobally(el::ConfigurationType::Enabled, "true");
    defaultConf.setGlobally(el::ConfigurationType::Format, "[%level] %msg");
    defaultConf.set(el::Level::Verbose, el::ConfigurationType::Format, "[%level-%vlevel] %msg");
    defaultConf.setGlobally(el::ConfigurationType::ToFile, "false");
    defaultConf.setGlobally(el::ConfigurationType::ToStandardOutput, "true");
#ifdef ELPP_DISABLE_VERBOSE_LOGS
    defaultConf.set(el::Level::Verbose, el::ConfigurationType::ToStandardOutput, "false");
#endif
    el::Loggers::reconfigureAllLoggers(defaultConf);
    el::Loggers::setVerboseLevel(1);

    std::cout << "This is HQS version 2, built on " << __DATE__ << " " << __TIME__ << '\n';
#ifndef NDEBUG
    LOG(WARNING) << "This is the debug version!";
#endif

    // Declare command line options
    bool        fast_refutation = true;
    bool        fraig           = false;
    bool        rewrite         = true;
    bool        pure            = false;
    bool        preprocess      = true;
    std::string in_name;
    std::string sat_command     = "";
    std::string qbf_command     = "";

    boost::program_options::options_description options("Options");
    options.add_options()("help", "Show the available options")(
        "filter", boost::program_options::value<bool>(&fast_refutation)->default_value(fast_refutation),
        "Use UNSAT filter with given value of k (0=off)")(
        "fraig", boost::program_options::value<bool>(&fraig)->default_value(fraig), "Use FRAIGing for the AIGs")(
        "rewrite", boost::program_options::value<bool>(&rewrite)->default_value(rewrite), "Use rewriting for the AIGs")(
        "pure", boost::program_options::value<bool>(&pure)->default_value(pure), "Use syntactic pure checks")(
        "preprocess", boost::program_options::value<bool>(&preprocess)->default_value(preprocess),
            "Apply preprocessing using HQSpre")(
        "sat", boost::program_options::value<std::string>(&sat_command), "Use external SAT solver")(
        "qbf", boost::program_options::value<std::string>(&qbf_command), "Use external QBF solver");

    boost::program_options::options_description invisible("Invisible options");
    invisible.add_options()("file", boost::program_options::value<std::string>(&in_name), "Input file");
    boost::program_options::options_description all("All options");
    all.add(options).add(invisible);

    boost::program_options::positional_options_description p;
    p.add("file", -1);

    boost::program_options::variables_map vm;
    boost::program_options::store(
        boost::program_options::command_line_parser(argc, argv).options(all).positional(p).run(), vm);
    boost::program_options::notify(vm);

    // Check command line options for plausibility
    if (vm.count("help") != 0u) {
        std::cout << "Call with\n  " << argv[0] << " [options] <input file>\n\n" << options << "\n\n";
        return 0;
    }

    if (in_name.empty()) {
        LOG(ERROR) << "No filename provided.";
        return -1;
    }

    // Parse the input file and load it into the preprocessor
    hqspre::Formula formula;
    formula.settings().consistency_check = false;

    const std::string ext = in_name.substr(in_name.length() - 3, 3);

    VLOG(1) << "Solving file " << in_name << " ...";

    try {
        if (ext == ".gz" || ext == ".GZ") {
            std::ifstream in(in_name, std::ios_base::in | std::ios_base::binary);
            if (!in) {
                LOG(ERROR) << "Could not open input file '" << in_name << "'!";
                std::exit(-1);
            }
            boost::iostreams::filtering_istream in_filter;
            in_filter.push(boost::iostreams::gzip_decompressor());
            in_filter.push(in);
            in_filter >> formula;
            in.close();
        } else {
            std::ifstream in(in_name);
            if (!in) {
                LOG(ERROR) << "Could not open input file '" << in_name << "'!";
                std::exit(-1);
            }
            in >> formula;
            in.close();
        }

        formula.determineGates();
        formula.settings().bla              = false;
        formula.settings().ble              = false;

        if (formula.getGates().size() > 5) {
        	
            // First do full preprocessing on a copy of the formula without preserving gates
            hqspre::Formula formula2(formula);
            formula2.settings().bla              = false;
            formula2.settings().ble              = false;
            formula2.settings().pure_sat_timeout = 1000;
            formula2.settings().impl_chains      = 3;
            formula2.settings().max_substitution_cost = 250;
            formula2.settings().max_resolution_cost = 100;
            formula2.settings().vivify_fp        = true;
            formula2.settings().sat_decide       = (sat_command == "");

            if (preprocess) {
                formula2.preprocess();
                if (formula2.numUVars() == 0 && sat_command != "") { // use external SAT solver
                    std::string fname = std::string(std::tmpnam(NULL))+".dimacs";
                    std::ofstream out(fname);
                    out << "p cnf " << formula2.maxVarIndex() << ' ' << formula2.numClauses() << '\n';
                    formula2.writeClauses(out, nullptr);
                    out.close();
                    const auto command = sat_command + " " + fname;
                    const auto ret = std::system(command.c_str());
                    std::remove(fname.c_str());
                    VLOG(1) << "External SAT solver returned: " << ret << '\n';
                    if (ret == 5120) throw hqspre::UNSATException("External solver solved the formula");
                    else if (ret == 2560) throw hqspre::SATException("External solver solved the formula");
                } else if (formula2.isQBF() && qbf_command != "") { // use external QBF solver
                    std::string fname = std::string(std::tmpnam(NULL))+".qdimacs";
                    std::ofstream out(fname);
                    formula2.write(out);
                    out.close();
                    const auto command = qbf_command + " " + fname;
                    const auto ret = std::system(command.c_str());
                    std::remove(fname.c_str());
                    VLOG(1) << "External QBF solver returned: " << ret << '\n';
                    if (ret == 5120) throw hqspre::UNSATException("External solver solved the formula");
                    else if (ret == 2560) throw hqspre::SATException("External solver solved the formula");
//                } else if (qbf_command != "" && fast_refutation) {
//                    std::ofstream out("test.qdimacs");
//
//                    // Print the QBF prefix
//                    out << "p cnf " << formula2.maxVarIndex() << ' ' << formula2.numClauses() << '\n';
//                    const auto pre = createQBFPrefix(formula2);
//                    char last = 'x';
//                    for (hqspre::Variable var: pre) {
//                        char q = (formula2.isExistential(var) ? 'e' : 'a');
//                        if (q != last) {
//                            if (last != 'x') out << " 0\n";
//                            last = q;
//                            out << q;
//                        }
//                        out << ' ' << var;
//                    }
//                    out << " 0\n";
//                    // Print the clauses
//                    formula2.writeClauses(out, nullptr);
//                    out.close();
//
//                    // Solve the formula. If it is UNSAT, so is the DQBF.
//                    const auto command = qbf_command + " test.qdimacs";
//                    const auto ret = std::system(command.c_str());
//                    VLOG(1) << "External QBF solver (filter) returned: " << ret << '\n';
//                    if (ret == 5120) throw hqspre::UNSATException("External solver decided the formula");
                }
            }
            

            // Then do preprocessing, preserving gates.
            formula.settings().univ_expand      = 0;
            formula.settings().preserve_gates   = true;
            formula.settings().substitution     = false;
            formula.settings().rewrite          = false;
            formula.settings().resolution       = false;
            formula.settings().max_loops        = 20;
            formula.settings().pure_sat_timeout = 1000;
            formula.settings().enableFork       = false;
        }
        if (preprocess) {
            formula.preprocess();
        }
        formula.printStatistics();
    } catch (hqspre::SATException&) {
        formula.printStatistics();
        VLOG(1) << "Solved by preprocessor.";
        std::cout << "[RESULT] SAT\n";

        solveTime = std::chrono::high_resolution_clock::now();
        duration = solveTime - beginTime;
        VLOG(1) << "Solving Time: " << duration.count();

        return result_SAT;
    } catch (hqspre::UNSATException&) {
        formula.printStatistics();
        VLOG(1) << "Solved by preprocessor.";
        std::cout << "[RESULT] UNSAT\n";

        solveTime = std::chrono::high_resolution_clock::now();
        duration = solveTime - beginTime;
        VLOG(1) << "Solving Time: " << duration.count();

        return result_UNSAT;
    } catch (hqspre::FileFormatException& e) {
        std::cerr << "Invalid input file (" << e.what() << ")\n";
        exit(0);
    }




    // Preprocessing is finished. Transfer the formula to the solver.
    hqs::HenkinSolver solver(formula);
    solver.settings().filter           = fast_refutation; // && (qbf_command == "");
    solver.settings().fraig            = fraig;
    solver.settings().rewrite          = rewrite;
    solver.settings().unit_pure_checks = pure;

    const auto result = solver.solve();
    VLOG(1) << "STATISTICS:\n" << solver.statistics();
    switch (result) {
        case hqspre::TruthValue::TRUE:
            std::cout << "[RESULT] SAT\n";
            solveTime = std::chrono::high_resolution_clock::now();
            duration = solveTime - beginTime;
            VLOG(1) << "Solving Time: " << duration.count();

            return result_SAT;
            break;
        case hqspre::TruthValue::FALSE:
            std::cout << "[RESULT] UNSAT\n";

            solveTime = std::chrono::high_resolution_clock::now();
            duration = solveTime - beginTime;
            VLOG(1) << "Solving Time: " << duration.count();

            return result_UNSAT;
            break;
        default:
            std::cout << "[RESULT] UNKNOWN\n";
            return 0;
            break;
    }

    return 0;
}

/*
        if (formula.numUVars() == 0 && !settings.sat_decide) {
            std::ofstream out("test.dimacs");
            out << "p cnf " << formula.maxVarIndex() << ' ' << formula.numClauses() << '\n';
            formula.writeClauses(out, nullptr);
            out.close();
            const auto command = settings.sat_command + " -q test.dimacs";
            const auto ret = std::system(command.c_str());
            std::cout << "RETURNED: " << ret << '\n';
            if (ret == 5120) throw hqspre::UNSATException("External solver solved the formula");
            else if (ret == 2560) throw hqspre::SATException("External solver solved the formula");
        }
*/
